@@ -2,8 +2,8 @@
 
   bignum.c -
 
-  $Author: nobu $
-  $Date: 2003/12/22 08:23:54 $
+  $Author: matz $
+  $Date: 2006/04/09 16:10:36 $
   created at: Fri Jun 10 00:48:55 JST 1994
 
   Copyright (C) 1993-2003 Yukihiro Matsumoto
@@ -42,11 +42,11 @@ static VALUE
 bignew_1(klass, len, sign)
     VALUE klass;
     long len;
-    char sign;
+    int sign;
 {
     NEWOBJ(big, struct RBignum);
     OBJSETUP(big, klass, T_BIGNUM);
-    big->sign = sign;
+    big->sign = sign?1:0;
     big->len = len;
     big->digits = ALLOC_N(BDIGIT, len);
 
@@ -65,10 +65,10 @@ rb_big_clone(x)
     return z;
 }
 
+/* modify a bignum by 2's complement */
 static void
-get2comp(x, carry)		/* get 2's complement */
+get2comp(x)
     VALUE x;
-    int carry;
 {
     long i = RBIGNUM(x)->len;
     BDIGIT *ds = BDIGITS(x);
@@ -81,11 +81,10 @@ get2comp(x, carry)		/* get 2's complement */
 	ds[i++] = BIGLO(num);
 	num = BIGDN(num);
     } while (i < RBIGNUM(x)->len);
-    if (!carry) return;
-    if ((ds[RBIGNUM(x)->len-1] & (1<<(BITSPERDIG-1))) == 0) {
+    if (num != 0) {
 	REALLOC_N(RBIGNUM(x)->digits, BDIGIT, ++RBIGNUM(x)->len);
 	ds = BDIGITS(x);
-	ds[RBIGNUM(x)->len-1] = ~0;
+	ds[RBIGNUM(x)->len-1] = RBIGNUM(x)->sign ? ~0 : 1;
     }
 }
 
@@ -93,7 +92,7 @@ void
 rb_big_2comp(x)			/* get 2's complement */
     VALUE x;
 {
-    get2comp(x, Qtrue);
+    get2comp(x);
 }
 
 static VALUE
@@ -203,6 +202,8 @@ rb_quad_pack(buf, val)
 	long len = RBIGNUM(val)->len;
 	BDIGIT *ds;
 
+	if (len > SIZEOF_LONG_LONG/SIZEOF_BDIGITS)
+	    rb_raise(rb_eRangeError, "bignum too big to convert into `quad int'");
 	ds = BDIGITS(val);
 	q = 0;
 	while (len--) {
@@ -272,7 +273,9 @@ rb_quad_pack(buf, val)
 	val = rb_int2big(FIX2LONG(val));
     }
     len = RBIGNUM(val)->len * SIZEOF_BDIGITS;
-    if (len > QUAD_SIZE) len = QUAD_SIZE;
+    if (len > QUAD_SIZE) {
+	rb_raise(rb_eRangeError, "bignum too big to convert into `quad int'");
+    }
     memcpy(buf, (char*)BDIGITS(val), len);
     if (!RBIGNUM(val)->sign) {
 	len = QUAD_SIZE;
@@ -459,6 +462,9 @@ rb_cstr_to_inum(str, base, badcheck)
 	    }
 	    continue;
 	}
+	else if (!ISASCII(c)) {
+	    break;
+	}
 	else if (isdigit(c)) {
 	    c -= '0';
 	}
@@ -740,6 +746,17 @@ big2ulong(x, type)
     return num;
 }
 
+unsigned long
+rb_big2ulong_pack(x)   
+    VALUE x;  
+{   
+    unsigned long num = big2ulong(x, "unsigned long");
+    if (!RBIGNUM(x)->sign) {
+	return -num;
+    }
+    return num;
+}  
+
 unsigned long
 rb_big2ulong(x)
     VALUE x;
@@ -759,10 +776,10 @@ long
 rb_big2long(x)
     VALUE x;
 {
-    unsigned long num = big2ulong(x, "int");
+    unsigned long num = big2ulong(x, "long");
 
     if ((long)num < 0 && (RBIGNUM(x)->sign || (long)num != LONG_MIN)) {
-	rb_raise(rb_eRangeError, "bignum too big to convert into `int'");
+	rb_raise(rb_eRangeError, "bignum too big to convert into `long'");
     }
     if (!RBIGNUM(x)->sign) return -(long)num;
     return num;
@@ -958,10 +975,14 @@ rb_big_eq(x, y)
       case T_BIGNUM:
 	break;
       case T_FLOAT:
-	if (rb_big2dbl(x) == RFLOAT(y)->value)
-	    return Qtrue;
-	else
-	    return Qfalse;
+        {
+	    volatile double a, b;
+
+	    a = RFLOAT(y)->value;
+	    if (isnan(a)) return Qfalse;
+	    b = rb_big2dbl(x);
+	    return (a == b)?Qtrue:Qfalse;
+	}
       default:
 	return rb_equal(y, x);
     }
@@ -1028,13 +1049,15 @@ rb_big_neg(x)
     VALUE x;
 {
     VALUE z = rb_big_clone(x);
-    long i = RBIGNUM(x)->len;
-    BDIGIT *ds = BDIGITS(z);
+    long i;
+    BDIGIT *ds;
 
-    if (!RBIGNUM(x)->sign) get2comp(z, Qtrue);
+    if (!RBIGNUM(x)->sign) get2comp(z);
+    ds = BDIGITS(z);
+    i = RBIGNUM(x)->len;
     while (i--) ds[i] = ~ds[i];
-    if (RBIGNUM(x)->sign) get2comp(z, Qfalse);
     RBIGNUM(z)->sign = !RBIGNUM(z)->sign;
+    if (RBIGNUM(x)->sign) get2comp(z);
 
     return bignorm(z);
 }
@@ -1065,7 +1088,7 @@ bigsub(x, y)
 	}
     }
 
-    z = bignew(RBIGNUM(x)->len, (z == 0)?1:0);
+    z = bignew(RBIGNUM(x)->len, z==0);
     zds = BDIGITS(z);
 
     for (i = 0, num = 0; i < RBIGNUM(y)->len; i++) { 
@@ -1089,7 +1112,7 @@ bigsub(x, y)
 static VALUE
 bigadd(x, y, sign)
     VALUE x, y;
-    char sign;
+    int sign;
 {
     VALUE z;
     BDIGIT_DBL num;
@@ -1579,10 +1602,15 @@ rb_big_pow(x, y)
 	break;
 
       case T_FIXNUM:
-	yy = NUM2LONG(y);
+	yy = FIX2LONG(y);
 	if (yy > 0) {
 	    VALUE z = x;
 
+	    if (RBIGNUM(x)->len * SIZEOF_BDIGITS * yy > 1024*1024) {
+		rb_warn("in a**b, b may be too big");
+		d = (double)yy;
+		break;
+	    }
 	    for (;;) {
 		yy -= 1;
 		if (yy == 0) break;
@@ -1611,25 +1639,26 @@ rb_big_pow(x, y)
  */
 
 VALUE
-rb_big_and(x, y)
-    VALUE x, y;
+rb_big_and(xx, yy)
+    VALUE xx, yy;
 {
-    VALUE z;
+    volatile VALUE x, y, z;
     BDIGIT *ds1, *ds2, *zds;
     long i, l1, l2;
     char sign;
 
-    y = rb_to_int(y);
+    x = xx;
+    y = rb_to_int(yy);
     if (FIXNUM_P(y)) {
 	y = rb_int2big(FIX2LONG(y));
     }
     if (!RBIGNUM(y)->sign) {
 	y = rb_big_clone(y);
-	get2comp(y, Qtrue);
+	get2comp(y);
     }
     if (!RBIGNUM(x)->sign) {
 	x = rb_big_clone(x);
-	get2comp(x, Qtrue);
+	get2comp(x);
     }
     if (RBIGNUM(x)->len > RBIGNUM(y)->len) {
 	l1 = RBIGNUM(y)->len;
@@ -1654,7 +1683,7 @@ rb_big_and(x, y)
     for (; i<l2; i++) {
 	zds[i] = sign?0:ds2[i];
     }
-    if (!RBIGNUM(z)->sign) get2comp(z, Qfalse);
+    if (!RBIGNUM(z)->sign) get2comp(z);
     return bignorm(z);
 }
 
@@ -1666,26 +1695,26 @@ rb_big_and(x, y)
  */
 
 VALUE
-rb_big_or(x, y)
-    VALUE x, y;
+rb_big_or(xx, yy)
+    VALUE xx, yy;
 {
-    VALUE z;
+    volatile VALUE x, y, z;
     BDIGIT *ds1, *ds2, *zds;
     long i, l1, l2;
     char sign;
 
-    y = rb_to_int(y);
+    x = xx;
+    y = rb_to_int(yy);
     if (FIXNUM_P(y)) {
 	y = rb_int2big(FIX2LONG(y));
     }
-
     if (!RBIGNUM(y)->sign) {
 	y = rb_big_clone(y);
-	get2comp(y, Qtrue);
+	get2comp(y);
     }
     if (!RBIGNUM(x)->sign) {
 	x = rb_big_clone(x);
-	get2comp(x, Qtrue);
+	get2comp(x);
     }
     if (RBIGNUM(x)->len > RBIGNUM(y)->len) {
 	l1 = RBIGNUM(y)->len;
@@ -1710,7 +1739,7 @@ rb_big_or(x, y)
     for (; i<l2; i++) {
 	zds[i] = sign?ds2[i]:(BIGRAD-1);
     }
-    if (!RBIGNUM(z)->sign) get2comp(z, Qfalse);
+    if (!RBIGNUM(z)->sign) get2comp(z);
 
     return bignorm(z);
 }
@@ -1723,26 +1752,27 @@ rb_big_or(x, y)
  */
 
 VALUE
-rb_big_xor(x, y)
-    VALUE x, y;
+rb_big_xor(xx, yy)
+    VALUE xx, yy;
 {
+    volatile VALUE x, y;
     VALUE z;
     BDIGIT *ds1, *ds2, *zds;
     long i, l1, l2;
     char sign;
 
-    y = rb_to_int(y);
+    x = xx;
+    y = rb_to_int(yy);
     if (FIXNUM_P(y)) {
 	y = rb_int2big(FIX2LONG(y));
     }
-
     if (!RBIGNUM(y)->sign) {
 	y = rb_big_clone(y);
-	get2comp(y, Qtrue);
+	get2comp(y);
     }
     if (!RBIGNUM(x)->sign) {
 	x = rb_big_clone(x);
-	get2comp(x, Qtrue);
+	get2comp(x);
     }
     if (RBIGNUM(x)->len > RBIGNUM(y)->len) {
 	l1 = RBIGNUM(y)->len;
@@ -1769,7 +1799,7 @@ rb_big_xor(x, y)
     for (; i<l2; i++) {
 	zds[i] = sign?ds2[i]:~ds2[i];
     }
-    if (!RBIGNUM(z)->sign) get2comp(z, Qfalse);
+    if (!RBIGNUM(z)->sign) get2comp(z);
 
     return bignorm(z);
 }
@@ -1830,6 +1860,7 @@ rb_big_rshift(x, y)
     VALUE z;
     BDIGIT_DBL num = 0;
     long i, j;
+    volatile VALUE save_x;
 
     if (shift < 0) return rb_big_lshift(x, INT2FIX(-shift));
 
@@ -1840,8 +1871,8 @@ rb_big_rshift(x, y)
 	    return INT2FIX(-1);
     }
     if (!RBIGNUM(x)->sign) {
-	x = rb_big_clone(x);
-	get2comp(x, Qtrue);
+	save_x = x = rb_big_clone(x);
+	get2comp(x);
     }
     xds = BDIGITS(x);
     i = RBIGNUM(x)->len; j = i - s1;
@@ -1856,7 +1887,7 @@ rb_big_rshift(x, y)
 	num = BIGUP(xds[i]);
     }
     if (!RBIGNUM(x)->sign) {
-	get2comp(z, Qfalse);
+	get2comp(z);
     }
     return bignorm(z);
 }
@@ -1901,7 +1932,7 @@ rb_big_aref(x, y)
     if (!RBIGNUM(x)->sign) {
 	if (s1 >= RBIGNUM(x)->len) return INT2FIX(1);
 	x = rb_big_clone(x);
-	get2comp(x, Qtrue);
+	get2comp(x);
     }
     else {
 	if (s1 >= RBIGNUM(x)->len) return INT2FIX(0);
@@ -1912,6 +1943,13 @@ rb_big_aref(x, y)
     return INT2FIX(0);
 }
 
+/*
+ * call-seq:
+ *   big.hash   => fixnum
+ *
+ * Compute a hash based on the value of _big_.
+ */
+
 static VALUE
 rb_big_hash(x)
     VALUE x;
@@ -1937,8 +1975,11 @@ rb_big_coerce(x, y)
     if (FIXNUM_P(y)) {
 	return rb_assoc_new(rb_int2big(FIX2LONG(y)), x);
     }
+    else if (TYPE(y) == T_BIGNUM) {
+       return rb_assoc_new(y, x);
+    }
     else {
-	rb_raise(rb_eTypeError, "Can't coerce %s to Bignum",
+	rb_raise(rb_eTypeError, "can't coerce %s to Bignum",
 		 rb_obj_classname(y));
     }
     /* not reached */
@@ -1977,11 +2018,13 @@ rb_big_rand(max, rand_buf)
 	return rb_float_new(rand_buf[0]);
     }
     v = bignew(len,1);
+    len--;
+    BDIGITS(v)[len] = BDIGITS(max)[len] * rand_buf[len];    
     while (len--) {
 	BDIGITS(v)[len] = ((BDIGIT)~0) * rand_buf[len];
     }
 
-    return rb_big_modulo((VALUE)v, max);
+    return v;
 }
 
 /*
