@@ -2,8 +2,8 @@
 
   error.c -
 
-  $Author: dave $
-  $Date: 2003/12/21 07:28:53 $
+  $Author: matz $
+  $Date: 2006/07/20 07:04:13 $
   created at: Mon Aug  9 16:11:34 JST 1993
 
   Copyright (C) 1993-2003 Yukihiro Matsumoto
@@ -12,7 +12,6 @@
 
 #include "ruby.h"
 #include "env.h"
-#include "version.h"
 #include "st.h"
 
 #include <stdio.h>
@@ -23,6 +22,14 @@
 #include <varargs.h>
 #define va_init_list(a,b) va_start(a)
 #endif
+#ifdef HAVE_STDLIB_H
+#include <stdlib.h>
+#endif
+#ifndef EXIT_SUCCESS
+#define EXIT_SUCCESS 0
+#endif
+
+extern const char ruby_version[], ruby_release_date[], ruby_platform[];
 
 int ruby_nerrs;
 
@@ -162,6 +169,14 @@ rb_warning(fmt, va_alist)
     va_end(args);
 }
 
+/*
+ * call-seq:
+ *    warn(msg)   => nil
+ *
+ * Display the given message (followed by a newline) on STDERR unless
+ * warnings are disabled (for example with the <code>-W0</code> flag).
+ */
+
 static VALUE
 rb_warn_m(self, mesg)
     VALUE self, mesg;
@@ -194,7 +209,7 @@ rb_bug(fmt, va_alist)
 	vfprintf(out, fmt, args);
 	va_end(args);
 	fprintf(out, "\nruby %s (%s) [%s]\n\n",
-		RUBY_VERSION, RUBY_RELEASE_DATE, RUBY_PLATFORM);
+		ruby_version, ruby_release_date, ruby_platform);
     }
     abort();
 }
@@ -251,6 +266,9 @@ rb_check_type(x, t)
 		else if (FIXNUM_P(x)) {
 		    etype = "Fixnum";
 		}
+		else if (SYMBOL_P(x)) {
+		    etype = "Symbol";
+		}
 		else if (rb_special_const_p(x)) {
 		    etype = RSTRING(rb_obj_as_string(x))->ptr;
 		}
@@ -285,6 +303,7 @@ VALUE rb_eNoMethodError;
 VALUE rb_eSecurityError;
 VALUE rb_eNotImpError;
 VALUE rb_eNoMemError;
+VALUE rb_cNameErrorMesg;
 
 VALUE rb_eScriptError;
 VALUE rb_eSyntaxError;
@@ -334,10 +353,7 @@ exc_initialize(argc, argv, exc)
 {
     VALUE arg;
 
-    if (rb_scan_args(argc, argv, "01", &arg) == 1) {
-	VALUE mesg = arg;
-	StringValue(mesg);	/* ensure mesg can be converted to String */
-    }
+    rb_scan_args(argc, argv, "01", &arg);
     rb_iv_set(exc, "mesg", arg);
     rb_iv_set(exc, "bt", Qnil);
 
@@ -345,6 +361,8 @@ exc_initialize(argc, argv, exc)
 }
 
 /*
+ *  Document-method: exception
+ *
  *  call-seq:
  *     exc.exception(string) -> an_exception or exc
  *  
@@ -385,7 +403,7 @@ exc_to_s(exc)
 {
     VALUE mesg = rb_attr_get(exc, rb_intern("mesg"));
 
-    if (NIL_P(mesg)) return rb_class_path(CLASS_OF(exc));
+    if (NIL_P(mesg)) return rb_class_name(CLASS_OF(exc));
     if (OBJ_TAINTED(exc)) OBJ_TAINT(mesg);
     return mesg;
 }
@@ -424,11 +442,11 @@ exc_inspect(exc)
     klass = CLASS_OF(exc);
     exc = rb_obj_as_string(exc);
     if (RSTRING(exc)->len == 0) {
-	return rb_str_dup(rb_class_path(klass));
+	return rb_str_dup(rb_class_name(klass));
     }
 
     str = rb_str_buf_new2("#<");
-    klass = rb_class_path(klass);
+    klass = rb_class_name(klass);
     rb_str_buf_append(str, klass);
     rb_str_buf_cat(str, ": ", 2);
     rb_str_buf_append(str, exc);
@@ -501,7 +519,7 @@ check_backtrace(bt)
 
 /*
  *  call-seq:
- *     exc.set_backtrace(array_   =>  array
+ *     exc.set_backtrace(array)   =>  array
  *  
  *  Sets the backtrace information associated with <i>exc</i>. The
  *  argument must be an array of <code>String</code> objects in the
@@ -517,22 +535,37 @@ exc_set_backtrace(exc, bt)
     return rb_iv_set(exc, "bt", check_backtrace(bt));
 }
 
+/*
+ * call-seq:
+ *   SystemExit.new(status=0)   => system_exit
+ *
+ * Create a new +SystemExit+ exception with the given status.
+ */
+
 static VALUE
 exit_initialize(argc, argv, exc)
     int argc;
     VALUE *argv;
     VALUE exc;
 {
-    VALUE status = INT2NUM(0);
+    VALUE status = INT2FIX(EXIT_SUCCESS);
     if (argc > 0 && FIXNUM_P(argv[0])) {
 	status = *argv++;
 	--argc;
     }
-    exc_initialize(argc, argv, exc);
+    rb_call_super(argc, argv);
     rb_iv_set(exc, "status", status);
     return exc;
 }
 
+
+/*
+ * call-seq:
+ *   system_exit.status   => fixnum
+ *
+ * Return the status value associated with this system exit.
+ */
+
 static VALUE
 exit_status(exc)
     VALUE exc;
@@ -540,6 +573,24 @@ exit_status(exc)
     return rb_attr_get(exc, rb_intern("status"));
 }
 
+
+/*
+ * call-seq:
+ *   system_exit.success?  => true or false
+ *
+ * Returns +true+ if exiting successful, +false+ if not.
+ */
+
+static VALUE
+exit_success_p(exc)
+    VALUE exc;
+{
+    VALUE status = rb_attr_get(exc, rb_intern("status"));
+    if (NIL_P(status)) return Qtrue;
+    if (status == INT2FIX(EXIT_SUCCESS)) return Qtrue;
+    return Qfalse;
+}
+
 void
 #ifdef HAVE_STDARG_PROTOTYPES
 rb_name_error(ID id, const char *fmt, ...)
@@ -564,6 +615,15 @@ rb_name_error(id, fmt, va_alist)
     rb_exc_raise(exc);
 }
 
+/*
+ * call-seq:
+ *   NameError.new(msg [, name])  => name_error
+ *
+ * Construct a new NameError exception. If given the <i>name</i>
+ * parameter may subsequently be examined using the <code>NameError.name</code>
+ * method.
+ */
+
 static VALUE
 name_err_initialize(argc, argv, self)
     int argc;
@@ -573,11 +633,18 @@ name_err_initialize(argc, argv, self)
     VALUE name;
 
     name = (argc > 1) ? argv[--argc] : Qnil;
-    exc_initialize(argc, argv, self);
+    rb_call_super(argc, argv);
     rb_iv_set(self, "name", name);
     return self;
 }
 
+/*
+ *  call-seq:
+ *    name_error.name    =>  string or nil
+ *
+ *  Return the name associated with this NameError exception.
+ */
+
 static VALUE
 name_err_name(self)
     VALUE self;
@@ -585,6 +652,38 @@ name_err_name(self)
     return rb_attr_get(self, rb_intern("name"));
 }
 
+/*
+ * call-seq:
+ *  name_error.to_s   => string
+ *
+ * Produce a nicely-formated string representing the +NameError+.
+ */
+
+static VALUE
+name_err_to_s(exc)
+    VALUE exc;
+{
+    VALUE mesg = rb_attr_get(exc, rb_intern("mesg")), str = mesg;
+
+    if (NIL_P(mesg)) return rb_class_name(CLASS_OF(exc));
+    StringValue(str);
+    if (str != mesg) {
+	rb_iv_set(exc, "mesg", mesg = str);
+    }
+    if (OBJ_TAINTED(exc)) OBJ_TAINT(mesg);
+    return mesg;
+}
+
+/*
+ * call-seq:
+ *   NoMethodError.new(msg, name [, args])  => no_method_error
+ *
+ * Construct a NoMethodError exception for a method of the given name
+ * called with the given arguments. The name may be accessed using
+ * the <code>#name</code> method on the resulting object, and the
+ * arguments using the <code>#args</code> method.
+ */
+
 static VALUE
 nometh_err_initialize(argc, argv, self)
     int argc;
@@ -597,6 +696,90 @@ nometh_err_initialize(argc, argv, self)
     return self;
 }
 
+/* :nodoc: */
+static void
+name_err_mesg_mark(ptr)
+    VALUE *ptr;
+{
+    rb_gc_mark_locations(ptr, ptr+3);
+}
+
+/* :nodoc: */
+static VALUE
+name_err_mesg_new(obj, mesg, recv, method)
+    VALUE obj, mesg, recv, method;
+{
+    VALUE *ptr = ALLOC_N(VALUE, 3);
+
+    ptr[0] = mesg;
+    ptr[1] = recv;
+    ptr[2] = method;
+    return Data_Wrap_Struct(rb_cNameErrorMesg, name_err_mesg_mark, -1, ptr);
+}
+
+/* :nodoc: */
+static VALUE
+name_err_mesg_to_str(obj)
+    VALUE obj;
+{
+    VALUE *ptr, mesg;
+    Data_Get_Struct(obj, VALUE, ptr);
+
+    mesg = ptr[0];
+    if (NIL_P(mesg)) return Qnil;
+    else {
+	char *desc = 0;
+	VALUE d = 0, args[3];
+
+	obj = ptr[1];
+	switch (TYPE(obj)) {
+	  case T_NIL:
+	    desc = "nil";
+	    break;
+	  case T_TRUE:
+	    desc = "true";
+	    break;
+	  case T_FALSE:
+	    desc = "false";
+	    break;
+	  default:
+	    d = rb_protect(rb_inspect, obj, 0);
+	    if (NIL_P(d) || RSTRING(d)->len > 65) {
+		d = rb_any_to_s(obj);
+	    }
+	    desc = RSTRING(d)->ptr;
+	    break;
+	}
+	if (desc && desc[0] != '#') {
+	    d = rb_str_new2(desc);
+	    rb_str_cat2(d, ":");
+	    rb_str_cat2(d, rb_obj_classname(obj));
+	}
+	args[0] = mesg;
+	args[1] = ptr[2];
+	args[2] = d;
+	mesg = rb_f_sprintf(3, args);
+    }
+    if (OBJ_TAINTED(obj)) OBJ_TAINT(mesg);
+    return mesg;
+}
+
+/* :nodoc: */
+static VALUE
+name_err_mesg_load(klass, str)
+    VALUE klass, str;
+{
+    return str;
+}
+
+/*
+ * call-seq:
+ *   no_method_error.args  => obj
+ *
+ * Return the arguments passed in as the third parameter to
+ * the constructor.
+ */
+
 static VALUE
 nometh_err_args(self)
     VALUE self;
@@ -679,6 +862,17 @@ get_syserr(n)
     return error;
 }
 
+/*
+ * call-seq:
+ *   SystemCallError.new(msg, errno)  => system_call_error_subclass
+ *
+ * If _errno_ corresponds to a known system error code, constructs
+ * the appropriate <code>Errno</code> class for that error, otherwise
+ * constructs a generic <code>SystemCallError</code> object. The
+ * error number is subsequently available via the <code>errno</code>
+ * method.
+ */
+
 static VALUE
 syserr_initialize(argc, argv, self)
     int argc;
@@ -689,7 +883,6 @@ syserr_initialize(argc, argv, self)
     char *strerror();
 #endif
     char *err;
-    char *buf;
     VALUE mesg, error;
     VALUE klass = rb_obj_class(self);
 
@@ -708,24 +901,36 @@ syserr_initialize(argc, argv, self)
     }
     else {
 	rb_scan_args(argc, argv, "01", &mesg);
-	error = rb_const_get_at(klass, rb_intern("Errno"));
+	error = rb_const_get(klass, rb_intern("Errno"));
     }
     if (!NIL_P(error)) err = strerror(NUM2LONG(error));
     else err = "unknown error";
     if (!NIL_P(mesg)) {
-	StringValue(mesg);
-	buf = ALLOCA_N(char, strlen(err)+RSTRING(mesg)->len+4);
-	sprintf(buf, "%s - %.*s", err, (int)RSTRING(mesg)->len, RSTRING(mesg)->ptr);
-	mesg = rb_str_new2(buf);
+	VALUE str = mesg;
+	size_t len;
+
+	StringValue(str);
+	len = strlen(err)+RSTRING(str)->len+3;
+	mesg = rb_str_new(0, len);
+	snprintf(RSTRING(mesg)->ptr, len+1, "%s - %.*s", err,
+		(int)RSTRING(str)->len, RSTRING(str)->ptr);
+	rb_str_resize(mesg, strlen(RSTRING(mesg)->ptr));
     }
     else {
 	mesg = rb_str_new2(err);
     }
-    exc_initialize(1, &mesg, self);
+    rb_call_super(1, &mesg);
     rb_iv_set(self, "errno", error);
     return self;
 }
 
+/*
+ * call-seq:
+ *   system_call_error.errno   => fixnum
+ *
+ * Return this SystemCallError's error number.
+ */
+
 static VALUE
 syserr_errno(self)
     VALUE self;
@@ -733,6 +938,15 @@ syserr_errno(self)
     return rb_attr_get(self, rb_intern("errno"));
 }
 
+/*
+ * call-seq:
+ *   system_call_error === other  => true or false
+ *
+ * Return +true+ if the receiver is a generic +SystemCallError+, or
+ * if the error numbers _self_ and _other_ are the same.
+ */
+
+
 static VALUE
 syserr_eqq(self, exc)
     VALUE self, exc;
@@ -784,6 +998,7 @@ Init_Exception()
     rb_eSystemExit  = rb_define_class("SystemExit", rb_eException);
     rb_define_method(rb_eSystemExit, "initialize", exit_initialize, -1);
     rb_define_method(rb_eSystemExit, "status", exit_status, 0);
+    rb_define_method(rb_eSystemExit, "success?", exit_success_p, 0);
 
     rb_eFatal  	    = rb_define_class("fatal", rb_eException);
     rb_eSignal      = rb_define_class("SignalException", rb_eException);
@@ -797,6 +1012,12 @@ Init_Exception()
     rb_eNameError     = rb_define_class("NameError", rb_eStandardError);
     rb_define_method(rb_eNameError, "initialize", name_err_initialize, -1);
     rb_define_method(rb_eNameError, "name", name_err_name, 0);
+    rb_define_method(rb_eNameError, "to_s", name_err_to_s, 0);
+    rb_cNameErrorMesg = rb_define_class_under(rb_eNameError, "message", rb_cData);
+    rb_define_singleton_method(rb_cNameErrorMesg, "!", name_err_mesg_new, 3);
+    rb_define_method(rb_cNameErrorMesg, "to_str", name_err_mesg_to_str, 0);
+    rb_define_method(rb_cNameErrorMesg, "_dump", name_err_mesg_to_str, 1);
+    rb_define_singleton_method(rb_cNameErrorMesg, "_load", name_err_mesg_load, 1);
     rb_eNoMethodError = rb_define_class("NoMethodError", rb_eNameError);
     rb_define_method(rb_eNoMethodError, "initialize", nometh_err_initialize, -1);
     rb_define_method(rb_eNoMethodError, "args", nometh_err_args, 0);
@@ -862,7 +1083,7 @@ void
 rb_notimplement()
 {
     rb_raise(rb_eNotImpError,
-	     "The %s() function is unimplemented on this machine",
+	     "the %s() function is unimplemented on this machine",
 	     rb_id2name(ruby_frame->last_func));
 }
 
@@ -890,7 +1111,6 @@ void
 rb_sys_fail(mesg)
     const char *mesg;
 {
-    extern int errno;
     int n = errno;
     VALUE arg;
 
@@ -931,14 +1151,14 @@ rb_sys_warning(fmt, va_alist)
 
 void
 rb_load_fail(path)
-    char *path;
+    const char *path;
 {
     rb_loaderror("%s -- %s", strerror(errno), path);
 }
 
 void
 rb_error_frozen(what)
-    char *what;
+    const char *what;
 {
     rb_raise(rb_eTypeError, "can't modify frozen %s", what);
 }
