@@ -2,8 +2,8 @@
 
   etc.c -
 
-  $Author: eban $
-  $Date: 2003/11/24 10:42:17 $
+  $Author: nobu $
+  $Date: 2006/06/27 02:37:01 $
   created at: Tue Mar 22 18:39:19 JST 1994
 
 ************************************************/
@@ -23,6 +23,10 @@
 #include <grp.h>
 #endif
 
+#ifndef HAVE_TYPE_UID_T
+#define uid_t int
+#endif
+
 static VALUE sPasswd, sGroup;
 
 #ifndef _WIN32
@@ -30,6 +34,11 @@ char *getenv();
 #endif
 char *getlogin();
 
+/* Returns the short user name of the currently logged in user.
+ *
+ * e.g.
+ *   Etc.getlogin -> 'guest'
+ */
 static VALUE
 etc_getlogin(obj)
     VALUE obj;
@@ -70,21 +79,21 @@ setup_passwd(pwd)
 #ifdef HAVE_ST_PW_PASSWD
 			 safe_setup_str(pwd->pw_passwd),
 #endif
-			 INT2FIX(pwd->pw_uid),
-			 INT2FIX(pwd->pw_gid),
+			 PW_UID2VAL(pwd->pw_uid),
+			 PW_GID2VAL(pwd->pw_gid),
 #ifdef HAVE_ST_PW_GECOS
 			 safe_setup_str(pwd->pw_gecos),
 #endif
 			 safe_setup_str(pwd->pw_dir),
 			 safe_setup_str(pwd->pw_shell),
 #ifdef HAVE_ST_PW_CHANGE
-			 INT2FIX(pwd->pw_change),
+			 INT2NUM(pwd->pw_change),
 #endif
 #ifdef HAVE_ST_PW_QUOTA
-			 INT2FIX(pwd->pw_quota),
+			 INT2NUM(pwd->pw_quota),
 #endif
 #ifdef HAVE_ST_PW_AGE
-			 INT2FIX(pwd->pw_age),
+			 PW_AGE2VAL(pwd->pw_age),
 #endif
 #ifdef HAVE_ST_PW_CLASS
 			 safe_setup_str(pwd->pw_class),
@@ -93,13 +102,22 @@ setup_passwd(pwd)
 			 safe_setup_str(pwd->pw_comment),
 #endif
 #ifdef HAVE_ST_PW_EXPIRE
-			 INT2FIX(pwd->pw_expire),
+			 INT2NUM(pwd->pw_expire),
 #endif
 			 0		/*dummy*/
 	);
 }
 #endif
 
+/* Returns the /etc/passwd information for the user with specified integer
+ * user id (uid).
+ *
+ * The information is returned as a Struct::Passwd; see getpwent above for
+ * details.
+ *
+ * e.g.  * Etc.getpwuid(0) -> #<struct Struct::Passwd name="root",
+ * passwd="x", uid=0, gid=0, gecos="root",dir="/root", shell="/bin/bash">
+ */
 static VALUE
 etc_getpwuid(argc, argv, obj)
     int argc;
@@ -126,6 +144,14 @@ etc_getpwuid(argc, argv, obj)
 #endif
 }
 
+/* Returns the /etc/passwd information for the user with specified login name.
+ *
+ * The information is returned as a Struct::Passwd; see getpwent above for
+ * details.
+ *
+ * e.g.  * Etc.getpwnam('root') -> #<struct Struct::Passwd name="root",
+ * passwd="x", uid=0, gid=0, gecos="root",dir="/root", shell="/bin/bash">
+ */
 static VALUE
 etc_getpwnam(obj, nam)
     VALUE obj, nam;
@@ -165,6 +191,21 @@ passwd_iterate()
 }
 #endif
 
+/* Provides a convenient Ruby iterator which executes a block for each entry 
+ * in the /etc/passwd file.
+ *
+ * The code block is passed an Etc::Passwd struct; see getpwent above for 
+ * details.
+ *
+ * Example:
+ *
+ *     require 'etc'
+ *
+ *     Etc.passwd {|u|
+ *       puts u.name + " = " + u.gecos
+ *     }
+ *
+ */
 static VALUE
 etc_passwd(obj)
     VALUE obj;
@@ -187,6 +228,9 @@ etc_passwd(obj)
     return Qnil;
 }
 
+/* Resets the process of reading the /etc/passwd file, so that the next call
+ * to getpwent will return the first entry again.
+ */
 static VALUE
 etc_setpwent(obj)
     VALUE obj;
@@ -197,6 +241,9 @@ etc_setpwent(obj)
     return Qnil;
 }
 
+/* Ends the process of scanning through the /etc/passwd file begun with
+ * getpwent, and closes the file.
+ */
 static VALUE
 etc_endpwent(obj)
     VALUE obj;
@@ -207,6 +254,32 @@ etc_endpwent(obj)
     return Qnil;
 }
 
+/* Returns an entry from the /etc/passwd file. The first time it is called it
+ * opens the file and returns the first entry; each successive call returns 
+ * the next entry, or nil if the end of the file has been reached.
+ *
+ * To close the file when processing is complete, call endpwent.
+ *
+ * Each entry is returned as a Struct::Passwd:
+ *
+ * - Passwd#name contains the short login name of the user as a String.
+ *
+ * - Passwd#passwd contains the encrypted password of the user as a String.
+ *   an 'x' is returned if shadow passwords are in use. An '*' is returned
+ *   if the user cannot log in using a password.
+ *
+ * - Passwd#uid contains the integer user ID (uid) of the user.
+ *
+ * - Passwd#gid contains the integer group ID (gid) of the user's primary group.
+ *
+ * - Passwd#gecos contains a longer String description of the user, such as 
+ *   a full name. Some Unix systems provide structured information in the 
+ *   gecos field, but this is system-dependent.
+ *
+ * - Passwd#dir contains the path to the home directory of the user as a String.
+ *
+ * - Passwd#shell contains the path to the login shell of the user as a String.
+ */
 static VALUE
 etc_getpwent(obj)
     VALUE obj;
@@ -240,11 +313,21 @@ setup_group(grp)
 #ifdef HAVE_ST_GR_PASSWD
 			 safe_setup_str(grp->gr_passwd),
 #endif
-			 INT2FIX(grp->gr_gid),
+			 PW_GID2VAL(grp->gr_gid),
 			 mem);
 }
 #endif
 
+/* Returns information about the group with specified integer group id (gid), 
+ * as found in /etc/group.
+ *
+ * The information is returned as a Struct::Group; see getgrent above for
+ * details.
+ *
+ * e.g.  Etc.getgrgid(100) -> #<struct Struct::Group name="users", passwd="x",
+ * gid=100, mem=["meta", "root"]>
+ *
+ */
 static VALUE
 etc_getgrgid(obj, id)
     VALUE obj, id;
@@ -263,6 +346,16 @@ etc_getgrgid(obj, id)
 #endif
 }
 
+/* Returns information about the group with specified String name, as found 
+ * in /etc/group.
+ *
+ * The information is returned as a Struct::Group; see getgrent above for
+ * details.
+ *
+ * e.g.  Etc.getgrnam('users') -> #<struct Struct::Group name="users",
+ * passwd="x", gid=100, mem=["meta", "root"]>
+ *
+ */
 static VALUE
 etc_getgrnam(obj, nam)
     VALUE obj, nam;
@@ -303,6 +396,21 @@ group_iterate()
 }
 #endif
 
+/* Provides a convenient Ruby iterator which executes a block for each entry 
+ * in the /etc/group file.
+ *
+ * The code block is passed an Etc::Group struct; see getgrent above for 
+ * details.
+ *
+ * Example:
+ *
+ *     require 'etc'
+ *
+ *     Etc.group {|g|
+ *       puts g.name + ": " + g.mem.join(', ')
+ *     }
+ *
+ */
 static VALUE
 etc_group(obj)
     VALUE obj;
@@ -325,6 +433,9 @@ etc_group(obj)
     return Qnil;
 }
 
+/* Resets the process of reading the /etc/group file, so that the next call
+ * to getgrent will return the first entry again.
+ */
 static VALUE
 etc_setgrent(obj)
     VALUE obj;
@@ -335,6 +446,9 @@ etc_setgrent(obj)
     return Qnil;
 }
 
+/* Ends the process of scanning through the /etc/group file begun by 
+ * getgrent, and closes the file.
+ */
 static VALUE
 etc_endgrent(obj)
     VALUE obj;
@@ -345,6 +459,26 @@ etc_endgrent(obj)
     return Qnil;
 }
 
+/* Returns an entry from the /etc/group file. The first time it is called it
+ * opens the file and returns the first entry; each successive call returns 
+ * the next entry, or nil if the end of the file has been reached.
+ *
+ * To close the file when processing is complete, call endgrent.
+ *
+ * Each entry is returned as a Struct::Group:
+ *
+ * - Group#name contains the name of the group as a String.
+ *
+ * - Group#passwd contains the encrypted password as a String. An 'x' is
+ *   returned if password access to the group is not available; an empty 
+ *   string is returned if no password is needed to obtain membership of 
+ *   the group.
+ *
+ * - Group#gid contains the group's numeric ID as an integer.
+ *
+ * - Group#mem is an Array of Strings containing the short login names of the 
+ *   members of the group.
+ */
 static VALUE
 etc_getgrent(obj)
     VALUE obj;
@@ -361,6 +495,11 @@ etc_getgrent(obj)
 
 static VALUE mEtc;
 
+/* The etc module provides access to information from the /etc/passwd and
+ * /etc/group files on Linux and Unix systems.
+ *
+ * Documented by mathew <meta@pobox.com>.
+ */
 void
 Init_etc()
 {
@@ -382,6 +521,7 @@ Init_etc()
     rb_define_module_function(mEtc, "endgrent", etc_endgrent, 0);
     rb_define_module_function(mEtc, "getgrent", etc_getgrent, 0);
 
+    rb_global_variable(&sPasswd);
     sPasswd =  rb_struct_define("Passwd",
 				"name", "passwd", "uid", "gid",
 #ifdef HAVE_ST_PW_GECOS
@@ -407,10 +547,13 @@ Init_etc()
 				"expire",
 #endif
 				NULL);
-    rb_global_variable(&sPasswd);
 
 #ifdef HAVE_GETGRENT
-    sGroup = rb_struct_define("Group", "name", "passwd", "gid", "mem", NULL);
     rb_global_variable(&sGroup);
+    sGroup = rb_struct_define("Group", "name",
+#ifdef HAVE_ST_GR_PASSWD
+			      "passwd",
+#endif
+			      "gid", "mem", NULL);
 #endif
 }
