@@ -103,8 +103,6 @@ char	*MasterDevice = "/dev/pty%s",
 
 static char SlaveName[DEVICELEN];
 
-extern int errno;
-
 #ifndef HAVE_SETEUID
 # ifdef HAVE_SETREUID
 #  define seteuid(e)	setreuid(-1, (e))
@@ -128,7 +126,7 @@ echild_status(self)
 
 struct pty_info {
     int fd;
-    pid_t child_pid;
+    rb_pid_t child_pid;
     VALUE thread;
 };
 
@@ -306,40 +304,36 @@ pty_finalize_syswait(info)
     return Qnil;
 }
 
-#ifdef HAVE_OPENPTY
+static int
+get_device_once(master, slave, fail)
+    int *master, *slave, fail;
+{
+#if defined HAVE_OPENPTY
 /*
  * Use openpty(3) of 4.3BSD Reno and later,
  * or the same interface function.
  */
-static void
-getDevice(master,slave)
-    int	*master,*slave;
-{
     if (openpty(master, slave, SlaveName,
 		(struct termios *)0, (struct winsize *)0) == -1) {
+	if (!fail) return -1;
 	rb_raise(rb_eRuntimeError, "openpty() failed");
     }
-}
-#else /* HAVE_OPENPTY */
-#ifdef HAVE__GETPTY
-static void
-getDevice(master,slave)
-    int	*master,*slave;
-{
+
+    return 0;
+#elif defined HAVE__GETPTY
     char *name;
 
     if (!(name = _getpty(master, O_RDWR, 0622, 0))) {
+	if (!fail) return -1;
 	rb_raise(rb_eRuntimeError, "_getpty() failed");
     }
 
     *slave = open(name, O_RDWR);
     strcpy(SlaveName, name);
+
+    return 0;
 }
 #else /* HAVE__GETPTY */
-static void
-getDevice(master,slave)
-    int	*master,*slave;
-{
     int	 i,j;
 
 #ifdef HAVE_PTSNAME
@@ -364,7 +358,7 @@ getDevice(master,slave)
 				*master = i;
 				*slave = j;
 				strcpy(SlaveName, pn);
-				return;
+				return 0;
 #if defined I_PUSH && !defined linux
 			    }
 			}
@@ -375,7 +369,8 @@ getDevice(master,slave)
 	}
 	close(i);
     }
-    rb_raise(rb_eRuntimeError, "Cannot get Master/Slave device");
+    if (!fail) rb_raise(rb_eRuntimeError, "can't get Master/Slave device");
+    return -1;
 #else
     char **p;
     char MasterName[DEVICELEN];
@@ -389,16 +384,26 @@ getDevice(master,slave)
 		*slave = j;
 		chown(SlaveName, getuid(), getgid());
 		chmod(SlaveName, 0622);
-		return;
+		return 0;
 	    }
 	    close(i);
 	}
     }
-    rb_raise(rb_eRuntimeError, "Cannot get %s", SlaveName);
+    if (fail) rb_raise(rb_eRuntimeError, "can't get %s", SlaveName);
+    return -1;
+#endif
 #endif
 }
-#endif /* HAVE__GETPTY */
-#endif /* HAVE_OPENPTY */
+
+static void
+getDevice(master, slave)
+    int *master, *slave;
+{
+    if (get_device_once(master, slave, 0)) {
+	rb_gc();
+	get_device_once(master, slave, 1);
+    }
+}
 
 static void
 freeDevice()
