@@ -1,12 +1,23 @@
 /* readline.c -- GNU Readline module
    Copyright (C) 1997-2001  Shugo Maeda */
 
+#include "config.h"
 #include <errno.h>
 #include <stdio.h>
+#include <ctype.h>
+#include <string.h>
+#ifdef HAVE_READLINE_READLINE_H
 #include <readline/readline.h>
+#endif
+#ifdef HAVE_READLINE_HISTORY_H
 #include <readline/history.h>
+#endif
+#ifdef HAVE_EDITLINE_READLINE_H
+#include <editline/readline.h>
+#endif
 
 #include "ruby.h"
+#include "rubyio.h"
 #include "rubysig.h"
 
 #ifdef HAVE_UNISTD_H
@@ -19,13 +30,18 @@ static VALUE mReadline;
 
 #define COMPLETION_PROC "completion_proc"
 #define COMPLETION_CASE_FOLD "completion_case_fold"
+static ID completion_proc, completion_case_fold;
 
-#ifndef READLINE_42_OR_LATER
+#ifndef HAVE_RL_FILENAME_COMPLETION_FUNCTION
 # define rl_filename_completion_function filename_completion_function
 # define rl_username_completion_function username_completion_function
 # define rl_completion_matches completion_matches
 #endif
 
+static int readline_event(void);
+static char **readline_attempted_completion_function(const char *text,
+                                                     int start, int end);
+
 static int
 readline_event()
 {
@@ -44,22 +60,33 @@ readline_readline(argc, argv, self)
     char *prompt = NULL;
     char *buff;
     int status;
+    OpenFile *ofp, *ifp;
 
+    rb_secure(4);
     if (rb_scan_args(argc, argv, "02", &tmp, &add_hist) > 0) {
-	prompt = StringValuePtr(tmp);
+	SafeStringValue(tmp);
+	prompt = RSTRING(tmp)->ptr;
     }
 
     if (!isatty(0) && errno == EBADF) rb_raise(rb_eIOError, "stdin closed");
 
+    Check_Type(rb_stdout, T_FILE);
+    GetOpenFile(rb_stdout, ofp);
+    rl_outstream = GetWriteFile(ofp);
+    Check_Type(rb_stdin, T_FILE);
+    GetOpenFile(rb_stdin, ifp);
+    rl_instream = GetReadFile(ifp);
     buff = (char*)rb_protect((VALUE(*)_((VALUE)))readline, (VALUE)prompt,
                               &status);
     if (status) {
-#if defined READLINE_40_OR_LATER
+#if defined HAVE_RL_CLEANUP_AFTER_SIGNAL
         /* restore terminal mode and signal handler*/
         rl_cleanup_after_signal();
-#elif defined READLINE_21_OR_LATER
+#elif defined HAVE_RL_DEPREP_TERM_FUNCTION
         /* restore terminal mode */
-        (*rl_deprep_term_function)();
+	if (rl_deprep_term_function != NULL) /* NULL in libedit. [ruby-dev:29116] */
+	    (*rl_deprep_term_function)();
+	else
 #else
         rl_deprep_terminal();
 #endif
@@ -82,16 +109,18 @@ readline_s_set_completion_proc(self, proc)
     VALUE self;
     VALUE proc;
 {
+    rb_secure(4);
     if (!rb_respond_to(proc, rb_intern("call")))
-	rb_raise(rb_eArgError, "argument have to respond to `call'");
-    return rb_iv_set(mReadline, COMPLETION_PROC, proc);
+	rb_raise(rb_eArgError, "argument must respond to `call'");
+    return rb_ivar_set(mReadline, completion_proc, proc);
 }
 
 static VALUE
 readline_s_get_completion_proc(self)
     VALUE self;
 {
-    return rb_iv_get(mReadline, COMPLETION_PROC);
+    rb_secure(4);
+    return rb_attr_get(mReadline, completion_proc);
 }
 
 static VALUE
@@ -99,19 +128,21 @@ readline_s_set_completion_case_fold(self, val)
     VALUE self;
     VALUE val;
 {
-    return rb_iv_set(mReadline, COMPLETION_CASE_FOLD, val);
+    rb_secure(4);
+    return rb_ivar_set(mReadline, completion_case_fold, val);
 }
 
 static VALUE
 readline_s_get_completion_case_fold(self)
     VALUE self;
 {
-    return rb_iv_get(mReadline, COMPLETION_CASE_FOLD);
+    rb_secure(4);
+    return rb_attr_get(mReadline, completion_case_fold);
 }
 
 static char **
 readline_attempted_completion_function(text, start, end)
-    char *text;
+    const char *text;
     int start;
     int end;
 {
@@ -120,11 +151,13 @@ readline_attempted_completion_function(text, start, end)
     int case_fold;
     int i, matches;
 
-    proc = rb_iv_get(mReadline, COMPLETION_PROC);
+    proc = rb_attr_get(mReadline, completion_proc);
     if (NIL_P(proc))
 	return NULL;
+#ifdef HAVE_RL_ATTEMPTED_COMPLETION_OVER
     rl_attempted_completion_over = 1;
-    case_fold = RTEST(rb_iv_get(mReadline, COMPLETION_CASE_FOLD));
+#endif
+    case_fold = RTEST(rb_attr_get(mReadline, completion_case_fold));
     ary = rb_funcall(proc, rb_intern("call"), 1, rb_tainted_str_new2(text));
     if (TYPE(ary) != T_ARRAY)
 	ary = rb_Array(ary);
@@ -140,9 +173,9 @@ readline_attempted_completion_function(text, start, end)
     result[matches + 1] = NULL;
 
     if (matches == 1) {
-	result[0] = result[1];
-	result[1] = NULL;
-    } else {
+        result[0] = strdup(result[1]);
+    }
+    else {
 	register int i = 1;
 	int low = 100000;
 
@@ -178,62 +211,83 @@ static VALUE
 readline_s_vi_editing_mode(self)
     VALUE self;
 {
+#ifdef HAVE_RL_VI_EDITING_MODE
+    rb_secure(4);
     rl_vi_editing_mode(1,0);
     return Qnil;
+#else
+    rb_notimplement();
+    return Qnil; /* not reached */
+#endif /* HAVE_RL_VI_EDITING_MODE */
 }
 
 static VALUE
 readline_s_emacs_editing_mode(self)
     VALUE self;
 {
+#ifdef HAVE_RL_EMACS_EDITING_MODE
+    rb_secure(4);
     rl_emacs_editing_mode(1,0);
     return Qnil;
+#else
+    rb_notimplement();
+    return Qnil; /* not reached */
+#endif /* HAVE_RL_EMACS_EDITING_MODE */
 }
 
 static VALUE
 readline_s_set_completion_append_character(self, str)
     VALUE self, str;
 {
-#ifdef READLINE_21_OR_LATER
-    if (NIL_P(str) || !StringValuePtr(str) || !RSTRING(str)->len) {
+#ifdef HAVE_RL_COMPLETION_APPEND_CHARACTER
+    rb_secure(4);
+    if (NIL_P(str)) {
 	rl_completion_append_character = '\0';
-    } else {
-	rl_completion_append_character = RSTRING(str)->ptr[0];
     }
-
+    else {
+	SafeStringValue(str);
+	if (RSTRING(str)->len == 0) {
+	    rl_completion_append_character = '\0';
+	} else {
+	    rl_completion_append_character = RSTRING(str)->ptr[0];
+	}
+    }
     return self;
 #else
     rb_notimplement();
-#endif /* READLINE_21_OR_LATER */
+    return Qnil; /* not reached */
+#endif /* HAVE_RL_COMPLETION_APPEND_CHARACTER */
 }
 
 static VALUE
 readline_s_get_completion_append_character(self)
     VALUE self;
 {
-#ifdef READLINE_21_OR_LATER
+#ifdef HAVE_RL_COMPLETION_APPEND_CHARACTER
     VALUE str;
 
+    rb_secure(4);
     if (rl_completion_append_character == '\0')
 	return Qnil;
 
     str = rb_str_new("", 1);
     RSTRING(str)->ptr[0] = rl_completion_append_character;
-
     return str;
 #else
     rb_notimplement();
-#endif /* READLINE_21_OR_LATER */
+    return Qnil; /* not reached */
+#endif /* HAVE_RL_COMPLETION_APPEND_CHARACTER */
 }
 
 static VALUE
 readline_s_set_basic_word_break_characters(self, str)
     VALUE self, str;
 {
-#ifdef READLINE_21_OR_LATER
+#ifdef HAVE_RL_BASIC_WORD_BREAK_CHARACTERS
     static char *basic_word_break_characters = NULL;
 
-    StringValue(str);
+    rb_secure(4);
+    SafeStringValue(str);
     if (basic_word_break_characters == NULL) {
 	basic_word_break_characters =
 	    ALLOC_N(char, RSTRING(str)->len + 1);
@@ -245,34 +299,37 @@ readline_s_set_basic_word_break_characters(self, str)
 	    RSTRING(str)->ptr, RSTRING(str)->len);
     basic_word_break_characters[RSTRING(str)->len] = '\0';
     rl_basic_word_break_characters = basic_word_break_characters;
-
     return self;
 #else
     rb_notimplement();
-#endif /* READLINE_21_OR_LATER */
+    return Qnil; /* not reached */
+#endif /* HAVE_RL_BASIC_WORD_BREAK_CHARACTERS */
 }
 
 static VALUE
 readline_s_get_basic_word_break_characters(self, str)
     VALUE self, str;
 {
-#ifdef READLINE_21_OR_LATER
+#ifdef HAVE_RL_BASIC_WORD_BREAK_CHARACTERS
+    rb_secure(4);
     if (rl_basic_word_break_characters == NULL)
 	return Qnil;
-    return rb_str_new2(rl_basic_word_break_characters);
+    return rb_tainted_str_new2(rl_basic_word_break_characters);
 #else
     rb_notimplement();
-#endif /* READLINE_21_OR_LATER */
+    return Qnil; /* not reached */
+#endif /* HAVE_RL_BASIC_WORD_BREAK_CHARACTERS */
 }
 
 static VALUE
 readline_s_set_completer_word_break_characters(self, str)
     VALUE self, str;
 {
-#ifdef READLINE_21_OR_LATER
+#ifdef HAVE_RL_COMPLETER_WORD_BREAK_CHARACTERS
     static char *completer_word_break_characters = NULL;
 
-    StringValue(str);
+    rb_secure(4);
+    SafeStringValue(str);
     if (completer_word_break_characters == NULL) {
 	completer_word_break_characters =
 	    ALLOC_N(char, RSTRING(str)->len + 1);
@@ -284,34 +341,37 @@ readline_s_set_completer_word_break_characters(self, str)
 	    RSTRING(str)->ptr, RSTRING(str)->len);
     completer_word_break_characters[RSTRING(str)->len] = '\0';
     rl_completer_word_break_characters = completer_word_break_characters;
-
     return self;
 #else
     rb_notimplement();
-#endif /* READLINE_21_OR_LATER */
+    return Qnil; /* not reached */
+#endif /* HAVE_RL_COMPLETER_WORD_BREAK_CHARACTERS */
 }
 
 static VALUE
 readline_s_get_completer_word_break_characters(self, str)
     VALUE self, str;
 {
-#ifdef READLINE_21_OR_LATER
+#ifdef HAVE_RL_COMPLETER_WORD_BREAK_CHARACTERS
+    rb_secure(4);
     if (rl_completer_word_break_characters == NULL)
 	return Qnil;
-    return rb_str_new2(rl_completer_word_break_characters);
+    return rb_tainted_str_new2(rl_completer_word_break_characters);
 #else
     rb_notimplement();
-#endif /* READLINE_21_OR_LATER */
+    return Qnil; /* not reached */
+#endif /* HAVE_RL_COMPLETER_WORD_BREAK_CHARACTERS */
 }
 
 static VALUE
 readline_s_set_basic_quote_characters(self, str)
     VALUE self, str;
 {
-#ifdef READLINE_21_OR_LATER
+#ifdef HAVE_RL_BASIC_QUOTE_CHARACTERS
     static char *basic_quote_characters = NULL;
 
-    StringValue(str);
+    rb_secure(4);
+    SafeStringValue(str);
     if (basic_quote_characters == NULL) {
 	basic_quote_characters =
 	    ALLOC_N(char, RSTRING(str)->len + 1);
@@ -327,30 +387,34 @@ readline_s_set_basic_quote_characters(self, str)
     return self;
 #else
     rb_notimplement();
-#endif /* READLINE_21_OR_LATER */
+    return Qnil; /* not reached */
+#endif /* HAVE_RL_BASIC_QUOTE_CHARACTERS */
 }
 
 static VALUE
 readline_s_get_basic_quote_characters(self, str)
     VALUE self, str;
 {
-#ifdef READLINE_21_OR_LATER
+#ifdef HAVE_RL_BASIC_QUOTE_CHARACTERS
+    rb_secure(4);
     if (rl_basic_quote_characters == NULL)
 	return Qnil;
-    return rb_str_new2(rl_basic_quote_characters);
+    return rb_tainted_str_new2(rl_basic_quote_characters);
 #else
     rb_notimplement();
-#endif /* READLINE_21_OR_LATER */
+    return Qnil; /* not reached */
+#endif /* HAVE_RL_BASIC_QUOTE_CHARACTERS */
 }
 
 static VALUE
 readline_s_set_completer_quote_characters(self, str)
     VALUE self, str;
 {
-#ifdef READLINE_21_OR_LATER
+#ifdef HAVE_RL_COMPLETER_QUOTE_CHARACTERS
     static char *completer_quote_characters = NULL;
 
-    StringValue(str);
+    rb_secure(4);
+    SafeStringValue(str);
     if (completer_quote_characters == NULL) {
 	completer_quote_characters =
 	    ALLOC_N(char, RSTRING(str)->len + 1);
@@ -366,30 +430,34 @@ readline_s_set_completer_quote_characters(self, str)
     return self;
 #else
     rb_notimplement();
-#endif /* READLINE_21_OR_LATER */
+    return Qnil; /* not reached */
+#endif /* HAVE_RL_COMPLETER_QUOTE_CHARACTERS */
 }
 
 static VALUE
 readline_s_get_completer_quote_characters(self, str)
     VALUE self, str;
 {
-#ifdef READLINE_21_OR_LATER
+#ifdef HAVE_RL_COMPLETER_QUOTE_CHARACTERS
+    rb_secure(4);
     if (rl_completer_quote_characters == NULL)
 	return Qnil;
-    return rb_str_new2(rl_completer_quote_characters);
+    return rb_tainted_str_new2(rl_completer_quote_characters);
 #else
     rb_notimplement();
-#endif /* READLINE_21_OR_LATER */
+    return Qnil; /* not reached */
+#endif /* HAVE_RL_COMPLETER_QUOTE_CHARACTERS */
 }
 
 static VALUE
 readline_s_set_filename_quote_characters(self, str)
     VALUE self, str;
 {
-#ifdef READLINE_21_OR_LATER
+#ifdef HAVE_RL_FILENAME_QUOTE_CHARACTERS
     static char *filename_quote_characters = NULL;
 
-    StringValue(str);
+    rb_secure(4);
+    SafeStringValue(str);
     if (filename_quote_characters == NULL) {
 	filename_quote_characters =
 	    ALLOC_N(char, RSTRING(str)->len + 1);
@@ -405,37 +473,23 @@ readline_s_set_filename_quote_characters(self, str)
     return self;
 #else
     rb_notimplement();
-#endif /* READLINE_21_OR_LATER */
+    return Qnil; /* not reached */
+#endif /* HAVE_RL_FILENAME_QUOTE_CHARACTERS */
 }
 
 static VALUE
 readline_s_get_filename_quote_characters(self, str)
     VALUE self, str;
 {
-#ifdef READLINE_21_OR_LATER
+#ifdef HAVE_RL_FILENAME_QUOTE_CHARACTERS
+    rb_secure(4);
     if (rl_filename_quote_characters == NULL)
 	return Qnil;
-    return rb_str_new2(rl_filename_quote_characters);
+    return rb_tainted_str_new2(rl_filename_quote_characters);
 #else
     rb_notimplement();
-#endif /* READLINE_21_OR_LATER */
-}
-
-static VALUE
-rb_remove_history(index)
-    int index;
-{
-    HIST_ENTRY *entry;
-    VALUE val;
-
-    entry = remove_history(index);
-    if (entry) {
-        val = rb_tainted_str_new2(entry->line);
-        free(entry->line);
-        free(entry);
-        return val;
-    }
-    return Qnil;
+    return Qnil; /* not reached */
+#endif /* HAVE_RL_FILENAME_QUOTE_CHARACTERS */
 }
 
 static VALUE
@@ -450,18 +504,19 @@ hist_get(self, index)
     VALUE self;
     VALUE index;
 {
-    HISTORY_STATE *state;
+    HIST_ENTRY *entry;
     int i;
 
-    state = history_get_history_state();
+    rb_secure(4);
     i = NUM2INT(index);
     if (i < 0) {
-        i += state->length;
+        i += history_length;
     }
-    if (i < 0 || i > state->length - 1) {
-	rb_raise(rb_eIndexError, "Invalid index");
+    entry = history_get(history_base + i);
+    if (entry == NULL) {
+	rb_raise(rb_eIndexError, "invalid index");
     }
-    return rb_tainted_str_new2(state->entries[i]->line);
+    return rb_tainted_str_new2(entry->line);
 }
 
 static VALUE
@@ -470,20 +525,25 @@ hist_set(self, index, str)
     VALUE index;
     VALUE str;
 {
-    HISTORY_STATE *state;
-    VALUE s = str;
+#ifdef HAVE_REPLACE_HISTORY_ENTRY
+    HIST_ENTRY *entry;
     int i;
 
-    state = history_get_history_state();
+    rb_secure(4);
     i = NUM2INT(index);
+    SafeStringValue(str);
     if (i < 0) {
-        i += state->length;
+        i += history_length;
     }
-    if (i < 0 || i > state->length - 1) {
-	rb_raise(rb_eIndexError, "Invalid index");
+    entry = replace_history_entry(i, RSTRING(str)->ptr, NULL);
+    if (entry == NULL) {
+	rb_raise(rb_eIndexError, "invalid index");
     }
-    replace_history_entry(i, StringValuePtr(s), NULL);
     return str;
+#else
+    rb_notimplement();
+    return Qnil; /* not reached */
+#endif
 }
 
 static VALUE
@@ -491,7 +551,9 @@ hist_push(self, str)
     VALUE self;
     VALUE str;
 {
-    add_history(StringValuePtr(str));
+    rb_secure(4);
+    SafeStringValue(str);
+    add_history(RSTRING(str)->ptr);
     return self;
 }
 
@@ -503,22 +565,45 @@ hist_push_method(argc, argv, self)
 {
     VALUE str;
     
+    rb_secure(4);
     while (argc--) {
 	str = *argv++;
-	add_history(StringValuePtr(str));
+	SafeStringValue(str);
+	add_history(RSTRING(str)->ptr);
     }
     return self;
 }
 
+static VALUE
+rb_remove_history(index)
+    int index;
+{
+#ifdef HAVE_REMOVE_HISTORY
+    HIST_ENTRY *entry;
+    VALUE val;
+
+    rb_secure(4);
+    entry = remove_history(index);
+    if (entry) {
+        val = rb_tainted_str_new2(entry->line);
+        free(entry->line);
+        free(entry);
+        return val;
+    }
+    return Qnil;
+#else
+    rb_notimplement();
+    return Qnil; /* not reached */
+#endif
+}
+
 static VALUE
 hist_pop(self)
     VALUE self;
 {
-    HISTORY_STATE *state;
-
-    state = history_get_history_state();
-    if (state->length > 0) {
-	return rb_remove_history(state->length - 1);
+    rb_secure(4);
+    if (history_length > 0) {
+	return rb_remove_history(history_length - 1);
     } else {
 	return Qnil;
     }
@@ -528,10 +613,8 @@ static VALUE
 hist_shift(self)
     VALUE self;
 {
-    HISTORY_STATE *state;
-
-    state = history_get_history_state();
-    if (state->length > 0) {
+    rb_secure(4);
+    if (history_length > 0) {
 	return rb_remove_history(0);
     } else {
 	return Qnil;
@@ -542,12 +625,15 @@ static VALUE
 hist_each(self)
     VALUE self;
 {
-    HISTORY_STATE *state;
+    HIST_ENTRY *entry;
     int i;
 
-    state = history_get_history_state();
-    for (i = 0; i < state->length; i++) {
-	rb_yield(rb_tainted_str_new2(state->entries[i]->line));
+    rb_secure(4);
+    for (i = 0; i < history_length; i++) {
+        entry = history_get(history_base + i);
+        if (entry == NULL)
+            break;
+	rb_yield(rb_tainted_str_new2(entry->line));
     }
     return self;
 }
@@ -556,23 +642,16 @@ static VALUE
 hist_length(self)
     VALUE self;
 {
-    HISTORY_STATE *state;
-
-    state = history_get_history_state();
-    return INT2NUM(state->length);
+    rb_secure(4);
+    return INT2NUM(history_length);
 }
 
 static VALUE
 hist_empty_p(self)
     VALUE self;
 {
-    HISTORY_STATE *state;
-
-    state = history_get_history_state();
-    if (state->length == 0)
-	return Qtrue;
-    else
-	return Qfalse;
+    rb_secure(4);
+    return history_length == 0 ? Qtrue : Qfalse;
 }
 
 static VALUE
@@ -580,15 +659,14 @@ hist_delete_at(self, index)
     VALUE self;
     VALUE index;
 {
-    HISTORY_STATE *state;
     int i;
 
-    state = history_get_history_state();
+    rb_secure(4);
     i = NUM2INT(index);
     if (i < 0)
-        i += state->length;
-    if (i < 0 || i > state->length - 1) {
-	rb_raise(rb_eIndexError, "Invalid index");
+        i += history_length;
+    if (i < 0 || i > history_length - 1) {
+	rb_raise(rb_eIndexError, "invalid index");
     }
     return rb_remove_history(i);
 }
@@ -657,6 +735,9 @@ Init_readline()
 
     using_history();
 
+    completion_proc = rb_intern(COMPLETION_PROC);
+    completion_case_fold = rb_intern(COMPLETION_CASE_FOLD);
+
     mReadline = rb_define_module("Readline");
     rb_define_module_function(mReadline, "readline",
 			      readline_readline, -1);
@@ -709,7 +790,6 @@ Init_readline()
     rb_define_singleton_method(history,"each", hist_each, 0);
     rb_define_singleton_method(history,"length", hist_length, 0);
     rb_define_singleton_method(history,"size", hist_length, 0);
-
     rb_define_singleton_method(history,"empty?", hist_empty_p, 0);
     rb_define_singleton_method(history,"delete_at", hist_delete_at, 1);
     rb_define_const(mReadline, "HISTORY", history);
@@ -723,21 +803,18 @@ Init_readline()
     rb_define_singleton_method(ucomp, "call",
 			       username_completion_proc_call, 1);
     rb_define_const(mReadline, "USERNAME_COMPLETION_PROC", ucomp);
-#if defined READLINE_21_OR_LATER
+#if defined HAVE_RL_LIBRARY_VERSION
     rb_define_const(mReadline, "VERSION", rb_str_new2(rl_library_version));
 #else
     rb_define_const(mReadline, "VERSION",
                     rb_str_new2("2.0 or before version"));
 #endif
 
-#if defined READLINE_42_OR_LATER
-    rl_attempted_completion_function
-	= (rl_completion_func_t *)readline_attempted_completion_function;
-    rl_event_hook = (rl_hook_func_t *)readline_event;
-#else
-    rl_attempted_completion_function
-	= (CPPFunction *) readline_attempted_completion_function;
+    rl_attempted_completion_function = readline_attempted_completion_function;
+#ifdef HAVE_RL_EVENT_HOOK
     rl_event_hook = readline_event;
 #endif
+#ifdef HAVE_RL_CLEAR_SIGNALS
     rl_clear_signals();
+#endif
 }
