@@ -3,7 +3,7 @@
   gc.c -
 
   $Author: matz $
-  $Date: 2003/08/02 06:45:57 $
+  $Date: 2006/08/25 08:12:46 $
   created at: Tue Oct  5 09:44:46 JST 1993
 
   Copyright (C) 1993-2003 Yukihiro Matsumoto
@@ -32,9 +32,21 @@
 
 #ifdef __ia64__
 #include <ucontext.h>
+#if defined(__FreeBSD__)
+/*
+ * FreeBSD/ia64 currently does not have a way for a process to get the
+ * base address for the RSE backing store, so hardcode it.
+ */
+#define __libc_ia64_register_backing_store_base (4ULL<<61)
+#else
 #pragma weak __libc_ia64_register_backing_store_base
 extern unsigned long __libc_ia64_register_backing_store_base;
 #endif
+#endif
+
+#if defined _WIN32 || defined __CYGWIN__
+#include <windows.h>
+#endif
 
 void re_free_registers _((struct re_registers*));
 void rb_io_fptr_finalize _((struct OpenFile*));
@@ -54,9 +66,7 @@ void rb_io_fptr_finalize _((struct OpenFile*));
 # ifdef HAVE_ALLOCA_H
 #  include <alloca.h>
 # else
-#  ifdef _AIX
- #pragma alloca
-#  else
+#  ifndef _AIX
 #   ifndef alloca /* predefined by HP cc +Olibcalls */
 void *alloca ();
 #   endif
@@ -76,13 +86,14 @@ static unsigned long malloc_increase = 0;
 static unsigned long malloc_limit = GC_MALLOC_LIMIT;
 static void run_final();
 static VALUE nomem_error;
+static void garbage_collect();
 
 void
 rb_memerror()
 {
     static int recurse = 0;
 
-    if (recurse > 0 && rb_safe_level() < 4) {
+    if (!nomem_error || (recurse > 0 && rb_safe_level() < 4)) {
 	fprintf(stderr, "[FATAL] failed to allocate memory\n");
 	exit(1);
     }
@@ -103,11 +114,11 @@ ruby_xmalloc(size)
     malloc_increase += size;
 
     if (malloc_increase > malloc_limit) {
-	rb_gc();
+	garbage_collect();
     }
     RUBY_CRITICAL(mem = malloc(size));
     if (!mem) {
-	rb_gc();
+	garbage_collect();
 	RUBY_CRITICAL(mem = malloc(size));
 	if (!mem) {
 	    rb_memerror();
@@ -144,7 +155,7 @@ ruby_xrealloc(ptr, size)
     malloc_increase += size;
     RUBY_CRITICAL(mem = realloc(ptr, size));
     if (!mem) {
-	rb_gc();
+	garbage_collect();
 	RUBY_CRITICAL(mem = realloc(ptr, size));
 	if (!mem) {
 	    rb_memerror();
@@ -168,6 +179,20 @@ static int during_gc;
 static int need_call_final = 0;
 static st_table *finalizer_table = 0;
 
+
+/*
+ *  call-seq:
+ *     GC.enable    => true or false
+ *
+ *  Enables garbage collection, returning <code>true</code> if garbage
+ *  collection was previously disabled.
+ *
+ *     GC.disable   #=> false
+ *     GC.enable    #=> true
+ *     GC.enable    #=> false
+ *
+ */
+
 VALUE
 rb_gc_enable()
 {
@@ -177,6 +202,18 @@ rb_gc_enable()
     return old;
 }
 
+/*
+ *  call-seq:
+ *     GC.disable    => true or false
+ *
+ *  Disables garbage collection, returning <code>true</code> if garbage
+ *  collection was already disabled.
+ *
+ *     GC.disable   #=> false
+ *     GC.disable   #=> true
+ *
+ */
+
 VALUE
 rb_gc_disable()
 {
@@ -228,6 +265,8 @@ rb_gc_unregister_address(addr)
     }
 }
 
+#undef GC_DEBUG
+
 void
 rb_global_variable(var)
     VALUE *var;
@@ -235,6 +274,10 @@ rb_global_variable(var)
     rb_gc_register_address(var);
 }
 
+#if defined(_MSC_VER) || defined(__BORLANDC__) || defined(__CYGWIN__)
+#pragma pack(push, 1) /* magic for reducing sizeof(RVALUE): 24 -> 20 */
+#endif
+
 typedef struct RVALUE {
     union {
 	struct {
@@ -255,21 +298,32 @@ typedef struct RVALUE {
 	struct RFile   file;
 	struct RNode   node;
 	struct RMatch  match;
-	struct RVarmap varmap; 
+	struct RVarmap varmap;
 	struct SCOPE   scope;
     } as;
+#ifdef GC_DEBUG
+    char *file;
+    int   line;
+#endif
 } RVALUE;
 
+#if defined(_MSC_VER) || defined(__BORLANDC__) || defined(__CYGWIN__)
+#pragma pack(pop)
+#endif
+
 static RVALUE *freelist = 0;
 static RVALUE *deferred_final_list = 0;
 
 #define HEAPS_INCREMENT 10
-static RVALUE **heaps;
+static struct heaps_slot {
+    void *membase;
+    RVALUE *slot;
+    int limit;
+} *heaps;
 static int heaps_length = 0;
 static int heaps_used   = 0;
 
 #define HEAP_MIN_SLOTS 10000
-static int *heaps_limits;
 static int heap_slots = HEAP_MIN_SLOTS;
 
 #define FREE_MIN  4096
@@ -283,20 +337,24 @@ add_heap()
 
     if (heaps_used == heaps_length) {
 	/* Realloc heaps */
+	struct heaps_slot *p;
+	int length;
+
 	heaps_length += HEAPS_INCREMENT;
-	RUBY_CRITICAL(heaps = (heaps_used>0)?
-			(RVALUE**)realloc(heaps, heaps_length*sizeof(RVALUE*)):
-			(RVALUE**)malloc(heaps_length*sizeof(RVALUE*)));
-	if (heaps == 0) rb_memerror();
-	RUBY_CRITICAL(heaps_limits = (heaps_used>0)?
-			(int*)realloc(heaps_limits, heaps_length*sizeof(int)):
-			(int*)malloc(heaps_length*sizeof(int)));
-	if (heaps_limits == 0) rb_memerror();
+	length = heaps_length*sizeof(struct heaps_slot);
+	RUBY_CRITICAL(
+	    if (heaps_used > 0) {
+		p = (struct heaps_slot *)realloc(heaps, length);
+		if (p) heaps = p;
+	    }
+	    else {
+		p = heaps = (struct heaps_slot *)malloc(length);
+	    });
+	if (p == 0) rb_memerror();
     }
 
     for (;;) {
-	RUBY_CRITICAL(p = heaps[heaps_used] = (RVALUE*)malloc(sizeof(RVALUE)*heap_slots));
-	heaps_limits[heaps_used] = heap_slots;
+	RUBY_CRITICAL(p = (RVALUE*)malloc(sizeof(RVALUE)*(heap_slots+1)));
 	if (p == 0) {
 	    if (heap_slots == HEAP_MIN_SLOTS) {
 		rb_memerror();
@@ -304,6 +362,13 @@ add_heap()
 	    heap_slots = HEAP_MIN_SLOTS;
 	    continue;
 	}
+        heaps[heaps_used].membase = p;
+        if ((VALUE)p % sizeof(RVALUE) == 0)
+            heap_slots += 1;
+        else
+            p = (RVALUE*)((VALUE)p + sizeof(RVALUE) - ((VALUE)p % sizeof(RVALUE)));
+        heaps[heaps_used].slot = p;
+        heaps[heaps_used].limit = heap_slots;
 	break;
     }
     pend = p + heap_slots;
@@ -311,6 +376,7 @@ add_heap()
     if (himem < pend) himem = pend;
     heaps_used++;
     heap_slots *= 1.8;
+    if (heap_slots <= 0) heap_slots = HEAP_MIN_SLOTS;
 
     while (p < pend) {
 	p->as.free.flags = 0;
@@ -326,11 +392,15 @@ rb_newobj()
 {
     VALUE obj;
 
-    if (!freelist) rb_gc();
+    if (!freelist) garbage_collect();
 
     obj = (VALUE)freelist;
     freelist = freelist->as.free.next;
     MEMZERO((void*)obj, RVALUE, 1);
+#ifdef GC_DEBUG
+    RANY(obj)->file = ruby_sourcefile;
+    RANY(obj)->line = ruby_sourceline;
+#endif
     return obj;
 }
 
@@ -342,6 +412,7 @@ rb_data_object_alloc(klass, datap, dmark, dfree)
     RUBY_DATA_FUNC dfree;
 {
     NEWOBJ(data, struct RData);
+    if (klass) Check_Type(klass, T_CLASS);
     OBJSETUP(data, klass, T_DATA);
     data->data = datap;
     data->dfree = dfree;
@@ -353,28 +424,37 @@ rb_data_object_alloc(klass, datap, dmark, dfree)
 extern st_table *rb_class_tbl;
 VALUE *rb_gc_stack_start = 0;
 
+#ifdef DJGPP
+/* set stack size (http://www.delorie.com/djgpp/v2faq/faq15_9.html) */
+unsigned int _stklen = 0x180000; /* 1.5 kB */
+#endif
+
 #if defined(DJGPP) || defined(_WIN32_WCE)
 static unsigned int STACK_LEVEL_MAX = 65535;
-#else
-#ifdef __human68k__
-extern unsigned int _stacksize;
+#elif defined(__human68k__)
+unsigned int _stacksize = 262144;
 # define STACK_LEVEL_MAX (_stacksize - 4096)
 # undef HAVE_GETRLIMIT
-#else
-#ifdef HAVE_GETRLIMIT
+#elif defined(HAVE_GETRLIMIT)
 static unsigned int STACK_LEVEL_MAX = 655300;
 #else
 # define STACK_LEVEL_MAX 655300
 #endif
-#endif
-#endif
 
 #ifdef C_ALLOCA
 # define SET_STACK_END VALUE stack_end; alloca(0);
 # define STACK_END (&stack_end)
 #else
 # if defined(__GNUC__) && defined(USE_BUILTIN_FRAME_ADDRESS) && !defined(__ia64__)
-#  define  SET_STACK_END    VALUE *stack_end = __builtin_frame_address(0)
+#  if ( __GNUC__ == 3 && __GNUC_MINOR__ > 0 ) || __GNUC__ > 3
+__attribute__ ((noinline))
+#  endif
+static VALUE *
+stack_end_address(void)
+{
+    return (VALUE *)__builtin_frame_address(0);
+}
+#  define  SET_STACK_END    VALUE *stack_end = stack_end_address()
 # else
 #  define  SET_STACK_END    VALUE *stack_end = alloca(1)
 # endif
@@ -395,23 +475,25 @@ static unsigned int STACK_LEVEL_MAX = 655300;
 #elif STACK_GROW_DIRECTION < 0
 # define STACK_UPPER(x, a, b) b
 #else
+static int grow_direction;
 static int
-stack_growup_p(addr)
+stack_grow_direction(addr)
     VALUE *addr;
 {
     SET_STACK_END;
 
-    if (STACK_END > addr) return Qtrue;
-    return Qfalse;
+    if (STACK_END > addr) return grow_direction = 1;
+    return grow_direction = -1;
 }
+# define stack_growup_p(x) ((grow_direction ? grow_direction : stack_grow_direction(x)) > 0)
 # define STACK_UPPER(x, a, b) (stack_growup_p(x) ? a : b)
 #endif
 
-#define GC_WARTER_MARK 512
+#define GC_WATER_MARK 512
 
 #define CHECK_STACK(ret) do {\
     SET_STACK_END;\
-    (ret) = (STACK_LENGTH > STACK_LEVEL_MAX + GC_WARTER_MARK);\
+    (ret) = (STACK_LENGTH > STACK_LEVEL_MAX + GC_WATER_MARK);\
 } while (0)
 
 int
@@ -488,7 +570,8 @@ sweep_source_filename(key, value)
     }
 }
 
-static void rb_gc_mark_children _((VALUE ptr));
+static void gc_mark _((VALUE ptr, int lev));
+static void gc_mark_children _((VALUE ptr, int lev));
 
 static void
 gc_mark_all()
@@ -498,11 +581,11 @@ gc_mark_all()
 
     init_mark_stack();
     for (i = 0; i < heaps_used; i++) {
-	p = heaps[i]; pend = p + heaps_limits[i];
+	p = heaps[i].slot; pend = p + heaps[i].limit;
 	while (p < pend) {
 	    if ((p->as.basic.flags & FL_MARK) &&
 		(p->as.basic.flags != FL_MARK)) {
-		rb_gc_mark_children((VALUE)p);
+		gc_mark_children((VALUE)p, 0);
 	    }
 	    p++;
 	}
@@ -521,7 +604,7 @@ gc_mark_rest()
     init_mark_stack();
     while(p != tmp_arry){
 	p--;
-	rb_gc_mark_children(*p);
+	gc_mark_children(*p, 0);
     }
 }
 
@@ -534,12 +617,12 @@ is_pointer_to_heap(ptr)
     register long i;
 
     if (p < lomem || p > himem) return Qfalse;
+    if ((VALUE)p % sizeof(RVALUE) != 0) return Qfalse;
 
     /* check if p looks like a pointer */
     for (i=0; i < heaps_used; i++) {
-	heap_org = heaps[i];
-	if (heap_org <= p && p < heap_org + heaps_limits[i] &&
-	    ((((char*)p)-((char*)heap_org))%sizeof(RVALUE)) == 0)
+	heap_org = heaps[i].slot;
+	if (heap_org <= p && p < heap_org + heaps[i].limit)
 	    return Qtrue;
     }
     return Qfalse;
@@ -550,9 +633,11 @@ mark_locations_array(x, n)
     register VALUE *x;
     register long n;
 {
+    VALUE v;
     while (n--) {
-	if (is_pointer_to_heap((void *)*x)) {
-	    rb_gc_mark(*x);
+        v = *x;
+	if (is_pointer_to_heap((void *)v)) {
+	    gc_mark(v, 0);
 	}
 	x++;
     }
@@ -562,7 +647,6 @@ void
 rb_gc_mark_locations(start, end)
     VALUE *start, *end;
 {
-    VALUE *tmp;
     long n;
 
     n = end - start;
@@ -570,38 +654,56 @@ rb_gc_mark_locations(start, end)
 }
 
 static int
-mark_entry(key, value)
+mark_entry(key, value, lev)
     ID key;
     VALUE value;
+    int lev;
 {
-    rb_gc_mark(value);
+    gc_mark(value, lev);
     return ST_CONTINUE;
 }
 
+static void
+mark_tbl(tbl, lev)
+    st_table *tbl;
+    int lev;
+{
+    if (!tbl) return;
+    st_foreach(tbl, mark_entry, lev);
+}
+
 void
 rb_mark_tbl(tbl)
     st_table *tbl;
 {
-    if (!tbl) return;
-    st_foreach(tbl, mark_entry, 0);
+    mark_tbl(tbl, 0);
 }
 
 static int
-mark_keyvalue(key, value)
+mark_keyvalue(key, value, lev)
     VALUE key;
     VALUE value;
+    int lev;
 {
-    rb_gc_mark(key);
-    rb_gc_mark(value);
+    gc_mark(key, lev);
+    gc_mark(value, lev);
     return ST_CONTINUE;
 }
 
+static void
+mark_hash(tbl, lev)
+    st_table *tbl;
+    int lev;
+{
+    if (!tbl) return;
+    st_foreach(tbl, mark_keyvalue, lev);
+}
+
 void
 rb_mark_hash(tbl)
     st_table *tbl;
 {
-    if (!tbl) return;
-    st_foreach(tbl, mark_keyvalue, 0);
+    mark_hash(tbl, 0);
 }
 
 void
@@ -609,25 +711,26 @@ rb_gc_mark_maybe(obj)
     VALUE obj;
 {
     if (is_pointer_to_heap((void *)obj)) {
-	rb_gc_mark(obj);
+	gc_mark(obj, 0);
     }
 }
 
-void
-rb_gc_mark(ptr)
+#define GC_LEVEL_MAX 250
+
+static void
+gc_mark(ptr, lev)
     VALUE ptr;
+    int lev;
 {
-    int ret;
     register RVALUE *obj;
 
     obj = RANY(ptr);
     if (rb_special_const_p(ptr)) return; /* special const not marked */
     if (obj->as.basic.flags == 0) return;       /* free cell */
-    if (obj->as.basic.flags & FL_MARK) return;  /* already marked */ 
+    if (obj->as.basic.flags & FL_MARK) return;  /* already marked */
     obj->as.basic.flags |= FL_MARK;
 
-    CHECK_STACK(ret);
-    if (ret) {
+    if (lev > GC_LEVEL_MAX || (lev == 0 && ruby_stack_check())) {
 	if (!mark_stack_overflow) {
 	    if (mark_stack_ptr - mark_stack < MARK_STACK_MAX) {
 		*mark_stack_ptr = ptr;
@@ -639,12 +742,20 @@ rb_gc_mark(ptr)
 	}
 	return;
     }
-    rb_gc_mark_children(ptr);
+    gc_mark_children(ptr, lev+1);
+}
+
+void
+rb_gc_mark(ptr)
+    VALUE ptr;
+{
+    gc_mark(ptr, 0);
 }
 
 static void
-rb_gc_mark_children(ptr)
+gc_mark_children(ptr, lev)
     VALUE ptr;
+    int lev;
 {
     register RVALUE *obj = RANY(ptr);
 
@@ -654,7 +765,7 @@ rb_gc_mark_children(ptr)
     obj = RANY(ptr);
     if (rb_special_const_p(ptr)) return; /* special const not marked */
     if (obj->as.basic.flags == 0) return;       /* free cell */
-    if (obj->as.basic.flags & FL_MARK) return;  /* already marked */ 
+    if (obj->as.basic.flags & FL_MARK) return;  /* already marked */
     obj->as.basic.flags |= FL_MARK;
 
   marking:
@@ -680,7 +791,7 @@ rb_gc_mark_children(ptr)
 	  case NODE_RESCUE:
 	  case NODE_RESBODY:
 	  case NODE_CLASS:
-	    rb_gc_mark((VALUE)obj->as.node.u2.node);
+	    gc_mark((VALUE)obj->as.node.u2.node, lev);
 	    /* fall through */
 	  case NODE_BLOCK:	/* 1,3 */
 	  case NODE_ARRAY:
@@ -693,7 +804,7 @@ rb_gc_mark_children(ptr)
 	  case NODE_CALL:
 	  case NODE_DEFS:
 	  case NODE_OP_ASGN1:
-	    rb_gc_mark((VALUE)obj->as.node.u1.node);
+	    gc_mark((VALUE)obj->as.node.u1.node, lev);
 	    /* fall through */
 	  case NODE_SUPER:	/* 3 */
 	  case NODE_FCALL:
@@ -717,7 +828,10 @@ rb_gc_mark_children(ptr)
 	  case NODE_OP_ASGN_OR:
 	  case NODE_OP_ASGN_AND:
 	  case NODE_MODULE:
-	    rb_gc_mark((VALUE)obj->as.node.u1.node);
+	  case NODE_ALIAS:
+	  case NODE_VALIAS:
+	  case NODE_ARGS:
+	    gc_mark((VALUE)obj->as.node.u1.node, lev);
 	    /* fall through */
 	  case NODE_METHOD:	/* 2 */
 	  case NODE_NOT:
@@ -731,6 +845,7 @@ rb_gc_mark_children(ptr)
 	  case NODE_COLON3:
 	  case NODE_OPT_N:
 	  case NODE_EVSTR:
+	  case NODE_UNDEF:
 	    ptr = (VALUE)obj->as.node.u2.node;
 	    goto again;
 
@@ -745,8 +860,8 @@ rb_gc_mark_children(ptr)
 	  case NODE_NEXT:
 	  case NODE_YIELD:
 	  case NODE_COLON2:
-	  case NODE_ARGS:
 	  case NODE_SPLAT:
+	  case NODE_TO_ARY:
 	  case NODE_SVALUE:
 	    ptr = (VALUE)obj->as.node.u1.node;
 	    goto again;
@@ -754,7 +869,7 @@ rb_gc_mark_children(ptr)
 	  case NODE_SCOPE:	/* 2,3 */
 	  case NODE_BLOCK_PASS:
 	  case NODE_CDECL:
-	    rb_gc_mark((VALUE)obj->as.node.u3.node);
+	    gc_mark((VALUE)obj->as.node.u3.node, lev);
 	    ptr = (VALUE)obj->as.node.u2.node;
 	    goto again;
 
@@ -769,11 +884,8 @@ rb_gc_mark_children(ptr)
 	  case NODE_CVAR:
 	  case NODE_NTH_REF:
 	  case NODE_BACK_REF:
-	  case NODE_ALIAS:
-	  case NODE_VALIAS:
 	  case NODE_REDO:
 	  case NODE_RETRY:
-	  case NODE_UNDEF:
 	  case NODE_SELF:
 	  case NODE_NIL:
 	  case NODE_TRUE:
@@ -782,35 +894,33 @@ rb_gc_mark_children(ptr)
 	  case NODE_BLOCK_ARG:
 	  case NODE_POSTEXE:
 	    break;
-#ifdef C_ALLOCA
 	  case NODE_ALLOCA:
 	    mark_locations_array((VALUE*)obj->as.node.u1.value,
 				 obj->as.node.u3.cnt);
 	    ptr = (VALUE)obj->as.node.u2.node;
 	    goto again;
-#endif
 
 	  default:		/* unlisted NODE */
 	    if (is_pointer_to_heap(obj->as.node.u1.node)) {
-		rb_gc_mark((VALUE)obj->as.node.u1.node);
+		gc_mark((VALUE)obj->as.node.u1.node, lev);
 	    }
 	    if (is_pointer_to_heap(obj->as.node.u2.node)) {
-		rb_gc_mark((VALUE)obj->as.node.u2.node);
+		gc_mark((VALUE)obj->as.node.u2.node, lev);
 	    }
 	    if (is_pointer_to_heap(obj->as.node.u3.node)) {
-		rb_gc_mark((VALUE)obj->as.node.u3.node);
+		gc_mark((VALUE)obj->as.node.u3.node, lev);
 	    }
 	}
 	return;			/* no need to mark class. */
     }
 
-    rb_gc_mark(obj->as.basic.klass);
+    gc_mark(obj->as.basic.klass, lev);
     switch (obj->as.basic.flags & T_MASK) {
       case T_ICLASS:
       case T_CLASS:
       case T_MODULE:
-	rb_mark_tbl(obj->as.klass.m_tbl);
-	rb_mark_tbl(obj->as.klass.iv_tbl);
+	mark_tbl(obj->as.klass.m_tbl, lev);
+	mark_tbl(obj->as.klass.iv_tbl, lev);
 	ptr = obj->as.klass.super;
 	goto again;
 
@@ -824,13 +934,13 @@ rb_gc_mark_children(ptr)
 	    VALUE *ptr = obj->as.array.ptr;
 
 	    for (i=0; i < len; i++) {
-		rb_gc_mark(*ptr++);
+		gc_mark(*ptr++, lev);
 	    }
 	}
 	break;
 
       case T_HASH:
-	rb_mark_hash(obj->as.hash.tbl);
+	mark_hash(obj->as.hash.tbl, lev);
 	ptr = obj->as.hash.ifnone;
 	goto again;
 
@@ -847,7 +957,7 @@ rb_gc_mark_children(ptr)
 	break;
 
       case T_OBJECT:
-	rb_mark_tbl(obj->as.object.iv_tbl);
+	mark_tbl(obj->as.object.iv_tbl, lev);
 	break;
 
       case T_FILE:
@@ -865,7 +975,7 @@ rb_gc_mark_children(ptr)
 	break;
 
       case T_VARMAP:
-	rb_gc_mark(obj->as.varmap.val);
+	gc_mark(obj->as.varmap.val, lev);
 	ptr = (VALUE)obj->as.varmap.next;
 	goto again;
 
@@ -875,7 +985,7 @@ rb_gc_mark_children(ptr)
 	    VALUE *vars = &obj->as.scope.local_vars[-1];
 
 	    while (n--) {
-		rb_gc_mark(*vars++);
+		gc_mark(*vars++, lev);
 	    }
 	}
 	break;
@@ -886,7 +996,7 @@ rb_gc_mark_children(ptr)
 	    VALUE *ptr = obj->as.rstruct.ptr;
 
 	    while (len--) {
-		rb_gc_mark(*ptr++);
+		gc_mark(*ptr++, lev);
 	    }
 	}
 	break;
@@ -900,29 +1010,76 @@ rb_gc_mark_children(ptr)
 
 static void obj_free _((VALUE));
 
+static void
+finalize_list(p)
+    RVALUE *p;
+{
+    while (p) {
+	RVALUE *tmp = p->as.free.next;
+	run_final((VALUE)p);
+	if (!FL_TEST(p, FL_SINGLETON)) { /* not freeing page */
+	    p->as.free.flags = 0;
+	    p->as.free.next = freelist;
+	    freelist = p;
+	}
+	p = tmp;
+    }
+}
+
+static void
+free_unused_heaps()
+{
+    int i, j;
+
+    for (i = j = 1; j < heaps_used; i++) {
+	if (heaps[i].limit == 0) {
+	    free(heaps[i].membase);
+	    heaps_used--;
+	}
+	else {
+	    if (i != j) {
+		heaps[j] = heaps[i];
+	    }
+	    j++;
+	}
+    }
+}
+
+void rb_gc_abort_threads(void);
+
 static void
 gc_sweep()
 {
     RVALUE *p, *pend, *final_list;
     int freed = 0;
-    int i, j;
+    int i;
     unsigned long live = 0;
+    unsigned long free_min = 0;
+
+    for (i = 0; i < heaps_used; i++) {
+        free_min += heaps[i].limit;
+    }
+    free_min = free_min * 0.2;
+    if (free_min < FREE_MIN)
+        free_min = FREE_MIN;
 
     if (ruby_in_compile && ruby_parser_stack_on_heap()) {
 	/* should not reclaim nodes during compilation
            if yacc's semantic stack is not allocated on machine stack */
 	for (i = 0; i < heaps_used; i++) {
-	    p = heaps[i]; pend = p + heaps_limits[i];
+	    p = heaps[i].slot; pend = p + heaps[i].limit;
 	    while (p < pend) {
 		if (!(p->as.basic.flags&FL_MARK) && BUILTIN_TYPE(p) == T_NODE)
-		    rb_gc_mark((VALUE)p);
+		    gc_mark((VALUE)p, 0);
 		p++;
 	    }
 	}
     }
 
     mark_source_filename(ruby_sourcefile);
-    st_foreach(source_filenames, sweep_source_filename, 0);
+    if (source_filenames) {
+        st_foreach(source_filenames, sweep_source_filename, 0);
+    }
 
     freelist = 0;
     final_list = deferred_final_list;
@@ -932,7 +1089,7 @@ gc_sweep()
 	RVALUE *free = freelist;
 	RVALUE *final = final_list;
 
-	p = heaps[i]; pend = p + heaps_limits[i];
+	p = heaps[i].slot; pend = p + heaps[i].limit;
 	while (p < pend) {
 	    if (!(p->as.basic.flags & FL_MARK)) {
 		if (p->as.basic.flags) {
@@ -952,7 +1109,7 @@ gc_sweep()
 	    }
 	    else if (RBASIC(p)->flags == FL_MARK) {
 		/* objects to be finalized */
-		/* do notning remain marked */
+		/* do nothing remain marked */
 	    }
 	    else {
 		RBASIC(p)->flags &= ~FL_MARK;
@@ -960,12 +1117,12 @@ gc_sweep()
 	    }
 	    p++;
 	}
-	if (n == heaps_limits[i] && freed + n > FREE_MIN) {
+	if (n == heaps[i].limit && freed > free_min) {
 	    RVALUE *pp;
 
-	    heaps_limits[i] = 0;
+	    heaps[i].limit = 0;
 	    for (pp = final_list; pp != final; pp = pp->as.free.next) {
-		p->as.free.flags |= FL_SINGLETON; /* freeing page mark */
+		pp->as.free.flags |= FL_SINGLETON; /* freeing page mark */
 	    }
 	    freelist = free;	/* cancel this page from freelist */
 	}
@@ -973,46 +1130,22 @@ gc_sweep()
 	    freed += n;
 	}
     }
-    malloc_limit += (malloc_increase - malloc_limit) * (double)live / (live + freed);
-    if (malloc_limit < GC_MALLOC_LIMIT) malloc_limit = GC_MALLOC_LIMIT;
+    if (malloc_increase > malloc_limit) {
+	malloc_limit += (malloc_increase - malloc_limit) * (double)live / (live + freed);
+	if (malloc_limit < GC_MALLOC_LIMIT) malloc_limit = GC_MALLOC_LIMIT;
+    }
     malloc_increase = 0;
-    if (freed < FREE_MIN) {
+    if (freed < free_min) {
 	add_heap();
     }
     during_gc = 0;
 
     /* clear finalization list */
     if (final_list) {
-	RVALUE *tmp;
-
-	if (rb_prohibit_interrupt || ruby_in_compile) {
-	    deferred_final_list = final_list;
-	    return;
-	}
-
-	for (p = final_list; p; p = tmp) {
-	    tmp = p->as.free.next;
-	    run_final((VALUE)p);
-	    if (!FL_TEST(p, FL_SINGLETON)) { /* not freeing page */
-		p->as.free.flags = 0;
-		p->as.free.next = freelist;
-		freelist = p;
-	    }
-	}
-    }
-    for (i = j = 1; j < heaps_used; i++) {
-	if (heaps_limits[i] == 0) {
-	    free(heaps[i]);
-	    heaps_used--;
-	}
-	else {
-	    if (i != j) {
-		heaps[j] = heaps[i];
-		heaps_limits[j] = heaps_limits[i];
-	    }
-	    j++;
-	}
+	deferred_final_list = final_list;
+	return;
     }
+    free_unused_heaps();
 }
 
 void
@@ -1121,11 +1254,9 @@ obj_free(obj)
 		RUBY_CRITICAL(free(RANY(obj)->as.node.u1.tbl));
 	    }
 	    break;
-#ifdef C_ALLOCA
 	  case NODE_ALLOCA:
 	    RUBY_CRITICAL(free(RANY(obj)->as.node.u1.node));
 	    break;
-#endif
 	}
 	return;			/* no need to free iv_tbl */
 
@@ -1147,8 +1278,8 @@ obj_free(obj)
 	break;
 
       default:
-	rb_bug("gc_sweep(): unknown data type 0x%lx(%ld)", obj,
-	       RANY(obj)->as.basic.flags & T_MASK);
+	rb_bug("gc_sweep(): unknown data type 0x%lx(0x%lx)",
+	       RANY(obj)->as.basic.flags & T_MASK, obj);
     }
 }
 
@@ -1156,9 +1287,7 @@ void
 rb_gc_mark_frame(frame)
     struct FRAME *frame;
 {
-    mark_locations_array(frame->argv, frame->argc);
-    rb_gc_mark(frame->cbase);
-    rb_gc_mark((VALUE)frame->node);
+    gc_mark((VALUE)frame->node, 0);
 }
 
 #ifdef __GNUC__
@@ -1199,14 +1328,19 @@ int rb_setjmp (rb_jmp_buf);
 #endif /* __human68k__ or DJGPP */
 #endif /* __GNUC__ */
 
-void
-rb_gc()
+static void
+garbage_collect()
 {
     struct gc_list *list;
     struct FRAME * volatile frame; /* gcc 2.7.2.3 -O2 bug??  */
     jmp_buf save_regs_gc_mark;
     SET_STACK_END;
 
+#ifdef HAVE_NATIVETHREAD
+    if (!is_ruby_native_thread()) {
+	rb_bug("cross-thread violation on rb_gc()");
+    }
+#endif
     if (dont_gc || during_gc) {
 	if (!freelist) {
 	    add_heap();
@@ -1217,10 +1351,12 @@ rb_gc()
     during_gc++;
 
     init_mark_stack();
-    
+
+    gc_mark((VALUE)ruby_current_node, 0);
+
     /* mark frame stack */
     for (frame = ruby_frame; frame; frame = frame->prev) {
-	rb_gc_mark_frame(frame); 
+	rb_gc_mark_frame(frame);
 	if (frame->tmp) {
 	    struct FRAME *tmp = frame->tmp;
 	    while (tmp) {
@@ -1229,11 +1365,10 @@ rb_gc()
 	    }
 	}
     }
-    rb_gc_mark((VALUE)ruby_class);
-    rb_gc_mark((VALUE)ruby_scope);
-    rb_gc_mark((VALUE)ruby_dyna_vars);
+    gc_mark((VALUE)ruby_scope, 0);
+    gc_mark((VALUE)ruby_dyna_vars, 0);
     if (finalizer_table) {
-	rb_mark_tbl(finalizer_table);
+	mark_tbl(finalizer_table, 0);
     }
 
     FLUSH_REGISTER_WINDOWS;
@@ -1260,7 +1395,11 @@ rb_gc()
 	mark_locations_array((VALUE*)&ctx.uc_mcontext,
 			     ((size_t)(sizeof(VALUE)-1 + sizeof ctx.uc_mcontext)/sizeof(VALUE)));
 	bot = (VALUE*)__libc_ia64_register_backing_store_base;
+#if defined(__FreeBSD__)
+	top = (VALUE*)ctx.uc_mcontext.mc_special.bspstore;
+#else
 	top = (VALUE*)ctx.uc_mcontext.sc_ar_bsp;
+#endif
 	rb_gc_mark_locations(bot, top);
     }
 #endif
@@ -1284,19 +1423,40 @@ rb_gc()
     rb_mark_generic_ivar_tbl();
 
     rb_gc_mark_parser();
-    
+
     /* gc_mark objects whose marking are not completed*/
-    while (!MARK_STACK_EMPTY){
-	if (mark_stack_overflow){
-	    gc_mark_all();
+    do {
+	while (!MARK_STACK_EMPTY) {
+	    if (mark_stack_overflow){
+		gc_mark_all();
+	    }
+	    else {
+		gc_mark_rest();
+	    }
 	}
-	else {
-	    gc_mark_rest();
-	}
-    }
+	rb_gc_abort_threads();
+    } while (!MARK_STACK_EMPTY);
+
     gc_sweep();
 }
 
+void
+rb_gc()
+{
+    garbage_collect();
+    rb_gc_finalize_deferred();
+}
+
+/*
+ *  call-seq:
+ *     GC.start                     => nil
+ *     gc.garbage_collect           => nil
+ *     ObjectSpace.garbage_collect  => nil
+ *
+ *  Initiates garbage collection, unless manually disabled.
+ *
+ */
+
 VALUE
 rb_gc_start()
 {
@@ -1304,19 +1464,36 @@ rb_gc_start()
     return Qnil;
 }
 
+void
+ruby_set_stack_size(size)
+    size_t size;
+{
+#ifndef STACK_LEVEL_MAX
+    STACK_LEVEL_MAX = size / sizeof(VALUE);
+#endif
+}
+
 void
 Init_stack(addr)
     VALUE *addr;
 {
-#if defined(__human68k__)
-    extern void *_SEND;
-    rb_gc_stack_start = _SEND;
+#if defined(_WIN32) || defined(__CYGWIN__)
+    MEMORY_BASIC_INFORMATION m;
+    memset(&m, 0, sizeof(m));
+    VirtualQuery(&m, &m, sizeof(m));
+    rb_gc_stack_start =
+	STACK_UPPER((VALUE *)&m, (VALUE *)m.BaseAddress,
+		    (VALUE *)((char *)m.BaseAddress + m.RegionSize) - 1);
+#elif defined(STACK_END_ADDRESS)
+    extern void *STACK_END_ADDRESS;
+    rb_gc_stack_start = STACK_END_ADDRESS;
 #else
     if (!addr) addr = (VALUE *)&addr;
+    STACK_UPPER(&addr, addr, ++addr);
     if (rb_gc_stack_start) {
 	if (STACK_UPPER(&addr,
 			rb_gc_stack_start > addr,
-			rb_gc_stack_start < ++addr))
+			rb_gc_stack_start < addr))
 	    rb_gc_stack_start = addr;
 	return;
     }
@@ -1327,23 +1504,47 @@ Init_stack(addr)
 	struct rlimit rlim;
 
 	if (getrlimit(RLIMIT_STACK, &rlim) == 0) {
-	    double space = (double)rlim.rlim_cur*0.2;
+	    unsigned int space = rlim.rlim_cur/5;
 
 	    if (space > 1024*1024) space = 1024*1024;
 	    STACK_LEVEL_MAX = (rlim.rlim_cur - space) / sizeof(VALUE);
 	}
     }
-#ifdef __ia64__
-    /* ruby crashes on IA64 if compiled with optimizer on */
-    /* when if STACK_LEVEL_MAX is greater than this magic number */
-    /* I know this is a kludge.  I suspect optimizer bug */
-#define IA64_MAGIC_STACK_LIMIT 49152
-    if (STACK_LEVEL_MAX > IA64_MAGIC_STACK_LIMIT)
-	STACK_LEVEL_MAX = IA64_MAGIC_STACK_LIMIT;
-#endif
 #endif
 }
 
+
+/*
+ * Document-class: ObjectSpace
+ *
+ *  The <code>ObjectSpace</code> module contains a number of routines
+ *  that interact with the garbage collection facility and allow you to
+ *  traverse all living objects with an iterator.
+ *
+ *  <code>ObjectSpace</code> also provides support for object
+ *  finalizers, procs that will be called when a specific object is
+ *  about to be destroyed by garbage collection.
+ *
+ *     include ObjectSpace
+ *
+ *
+ *     a = "A"
+ *     b = "B"
+ *     c = "C"
+ *
+ *
+ *     define_finalizer(a, proc {|id| puts "Finalizer one on #{id}" })
+ *     define_finalizer(a, proc {|id| puts "Finalizer two on #{id}" })
+ *     define_finalizer(b, proc {|id| puts "Finalizer three on #{id}" })
+ *
+ *  <em>produces:</em>
+ *
+ *     Finalizer three on 537763470
+ *     Finalizer one on 537763480
+ *     Finalizer two on 537763480
+ *
+ */
+
 void
 Init_heap()
 {
@@ -1362,7 +1563,7 @@ os_live_obj()
     for (i = 0; i < heaps_used; i++) {
 	RVALUE *p, *pend;
 
-	p = heaps[i]; pend = p + heaps_limits[i];
+	p = heaps[i].slot; pend = p + heaps[i].limit;
 	for (;p < pend; p++) {
 	    if (p->as.basic.flags) {
 		switch (TYPE(p)) {
@@ -1395,7 +1596,7 @@ os_obj_of(of)
     for (i = 0; i < heaps_used; i++) {
 	RVALUE *p, *pend;
 
-	p = heaps[i]; pend = p + heaps_limits[i];
+	p = heaps[i].slot; pend = p + heaps[i].limit;
 	for (;p < pend; p++) {
 	    if (p->as.basic.flags) {
 		switch (TYPE(p)) {
@@ -1420,6 +1621,39 @@ os_obj_of(of)
     return INT2FIX(n);
 }
 
+/*
+ *  call-seq:
+ *     ObjectSpace.each_object([module]) {|obj| ... } => fixnum
+ *
+ *  Calls the block once for each living, nonimmediate object in this
+ *  Ruby process. If <i>module</i> is specified, calls the block
+ *  for only those classes or modules that match (or are a subclass of)
+ *  <i>module</i>. Returns the number of objects found. Immediate
+ *  objects (<code>Fixnum</code>s, <code>Symbol</code>s
+ *  <code>true</code>, <code>false</code>, and <code>nil</code>) are
+ *  never returned. In the example below, <code>each_object</code>
+ *  returns both the numbers we defined and several constants defined in
+ *  the <code>Math</code> module.
+ *
+ *     a = 102.7
+ *     b = 95       # Won't be returned
+ *     c = 12345678987654321
+ *     count = ObjectSpace.each_object(Numeric) {|x| p x }
+ *     puts "Total count: #{count}"
+ *
+ *  <em>produces:</em>
+ *
+ *     12345678987654321
+ *     102.7
+ *     2.71828182845905
+ *     3.14159265358979
+ *     2.22044604925031e-16
+ *     1.7976931348623157e+308
+ *     2.2250738585072e-308
+ *     Total count: 7
+ *
+ */
+
 static VALUE
 os_each_obj(argc, argv)
     int argc;
@@ -1438,6 +1672,9 @@ os_each_obj(argc, argv)
 
 static VALUE finalizers;
 
+/* deprecated
+ */
+
 static VALUE
 add_final(os, block)
     VALUE os, block;
@@ -1451,6 +1688,9 @@ add_final(os, block)
     return block;
 }
 
+/*
+ * deprecated
+ */
 static VALUE
 rm_final(os, block)
     VALUE os, block;
@@ -1460,6 +1700,9 @@ rm_final(os, block)
     return block;
 }
 
+/*
+ * deprecated
+ */
 static VALUE
 finals()
 {
@@ -1467,6 +1710,10 @@ finals()
     return finalizers;
 }
 
+/*
+ * deprecated
+ */
+
 static VALUE
 call_final(os, obj)
     VALUE os, obj;
@@ -1477,6 +1724,14 @@ call_final(os, obj)
     return obj;
 }
 
+/*
+ *  call-seq:
+ *     ObjectSpace.undefine_finalizer(obj)
+ *
+ *  Removes all finalizers for <i>obj</i>.
+ *
+ */
+
 static VALUE
 undefine_final(os, obj)
     VALUE os, obj;
@@ -1487,6 +1742,15 @@ undefine_final(os, obj)
     return obj;
 }
 
+/*
+ *  call-seq:
+ *     ObjectSpace.define_finalizer(obj, aProc=proc())
+ *
+ *  Adds <i>aProc</i> as a finalizer, to be called after <i>obj</i>
+ *  was destroyed.
+ *
+ */
+
 static VALUE
 define_final(argc, argv, os)
     int argc;
@@ -1506,6 +1770,8 @@ define_final(argc, argv, os)
     need_call_final = 1;
     FL_SET(obj, FL_FINALIZE);
 
+    block = rb_ary_new3(2, INT2FIX(ruby_safe_level), block);
+
     if (!finalizer_table) {
 	finalizer_table = st_init_numtable();
     }
@@ -1526,19 +1792,17 @@ rb_gc_copy_finalizer(dest, obj)
 
     if (!finalizer_table) return;
     if (!FL_TEST(obj, FL_FINALIZE)) return;
-    if (FL_TEST(dest, FL_FINALIZE)) {
-	rb_warn("copy_finalizer: descarding old finalizers");
-    }
     if (st_lookup(finalizer_table, obj, &table)) {
 	st_insert(finalizer_table, dest, table);
     }
+    RBASIC(dest)->flags |= FL_FINALIZE;
 }
 
 static VALUE
 run_single_final(args)
     VALUE *args;
 {
-    rb_eval_cmd(args[0], args[1], 0);
+    rb_eval_cmd(args[0], args[1], (int)args[2]);
     return Qnil;
 }
 
@@ -1548,23 +1812,41 @@ run_final(obj)
 {
     long i;
     int status, critical_save = rb_thread_critical;
-    VALUE args[2], table;
+    VALUE args[3], table, objid;
 
+    objid = rb_obj_id(obj);	/* make obj into id */
     rb_thread_critical = Qtrue;
-    args[1] = rb_ary_new3(1, rb_obj_id(obj)); /* make obj into id */
+    args[1] = 0;
+    args[2] = (VALUE)ruby_safe_level;
     for (i=0; i<RARRAY(finalizers)->len; i++) {
 	args[0] = RARRAY(finalizers)->ptr[i];
+	if (!args[1]) args[1] = rb_ary_new3(1, objid);
 	rb_protect((VALUE(*)_((VALUE)))run_single_final, (VALUE)args, &status);
     }
     if (finalizer_table && st_delete(finalizer_table, (st_data_t*)&obj, &table)) {
 	for (i=0; i<RARRAY(table)->len; i++) {
-	    args[0] = RARRAY(table)->ptr[i];
+	    VALUE final = RARRAY(table)->ptr[i];
+	    args[0] = RARRAY(final)->ptr[1];
+	    if (!args[1]) args[1] = rb_ary_new3(1, objid);
+	    args[2] = FIX2INT(RARRAY(final)->ptr[0]);
 	    rb_protect((VALUE(*)_((VALUE)))run_single_final, (VALUE)args, &status);
 	}
     }
     rb_thread_critical = critical_save;
 }
 
+void
+rb_gc_finalize_deferred()
+{
+    RVALUE *p = deferred_final_list;
+
+    deferred_final_list = 0;
+    if (p) {
+	finalize_list(p);
+	free_unused_heaps();
+    }
+}
+
 void
 rb_gc_call_finalizer_at_exit()
 {
@@ -1573,16 +1855,11 @@ rb_gc_call_finalizer_at_exit()
 
     /* run finalizers */
     if (need_call_final) {
-	if (deferred_final_list) {
-	    p = deferred_final_list;
-	    while (p) {
-		RVALUE *tmp = p;
-		p = p->as.free.next;
-		run_final((VALUE)tmp);
-	    }
-	}
+	p = deferred_final_list;
+	deferred_final_list = 0;
+	finalize_list(p);
 	for (i = 0; i < heaps_used; i++) {
-	    p = heaps[i]; pend = p + heaps_limits[i];
+	    p = heaps[i].slot; pend = p + heaps[i].limit;
 	    while (p < pend) {
 		if (FL_TEST(p, FL_FINALIZE)) {
 		    FL_UNSET(p, FL_FINALIZE);
@@ -1595,7 +1872,7 @@ rb_gc_call_finalizer_at_exit()
     }
     /* run data object's finalizers */
     for (i = 0; i < heaps_used; i++) {
-	p = heaps[i]; pend = p + heaps_limits[i];
+	p = heaps[i].slot; pend = p + heaps[i].limit;
 	while (p < pend) {
 	    if (BUILTIN_TYPE(p) == T_DATA &&
 		DATA_PTR(p) && RANY(p)->as.data.dfree) {
@@ -1616,32 +1893,122 @@ rb_gc_call_finalizer_at_exit()
     }
 }
 
+/*
+ *  call-seq:
+ *     ObjectSpace._id2ref(object_id) -> an_object
+ *
+ *  Converts an object id to a reference to the object. May not be
+ *  called on an object id passed as a parameter to a finalizer.
+ *
+ *     s = "I am a string"                    #=> "I am a string"
+ *     r = ObjectSpace._id2ref(s.object_id)   #=> "I am a string"
+ *     r == s                                 #=> true
+ *
+ */
+
 static VALUE
-id2ref(obj, id)
-    VALUE obj, id;
+id2ref(obj, objid)
+    VALUE obj, objid;
 {
     unsigned long ptr, p0;
 
     rb_secure(4);
-    p0 = ptr = NUM2ULONG(id);
+    p0 = ptr = NUM2ULONG(objid);
     if (ptr == Qtrue) return Qtrue;
     if (ptr == Qfalse) return Qfalse;
     if (ptr == Qnil) return Qnil;
     if (FIXNUM_P(ptr)) return (VALUE)ptr;
-    if (SYMBOL_P(ptr) && rb_id2name(SYM2ID((VALUE)ptr)) != 0) {
-	return (VALUE)ptr;
+    ptr = objid ^ FIXNUM_FLAG;	/* unset FIXNUM_FLAG */
+
+    if ((ptr % sizeof(RVALUE)) == (4 << 2)) {
+        ID symid = ptr / sizeof(RVALUE);
+        if (rb_id2name(symid) == 0)
+            rb_raise(rb_eRangeError, "%p is not symbol id value", p0);
+        return ID2SYM(symid);
     }
 
-    ptr = id ^ FIXNUM_FLAG;	/* unset FIXNUM_FLAG */
-    if (!is_pointer_to_heap((void *)ptr)) {
+    if (!is_pointer_to_heap((void *)ptr)|| BUILTIN_TYPE(ptr) >= T_BLKTAG) {
 	rb_raise(rb_eRangeError, "0x%lx is not id value", p0);
     }
-    if (BUILTIN_TYPE(ptr) == 0) {
+    if (BUILTIN_TYPE(ptr) == 0 || RBASIC(ptr)->klass == 0) {
 	rb_raise(rb_eRangeError, "0x%lx is recycled object", p0);
     }
     return (VALUE)ptr;
 }
 
+/*
+ *  Document-method: __id__
+ *  Document-method: object_id
+ *
+ *  call-seq:
+ *     obj.__id__       => fixnum
+ *     obj.object_id    => fixnum
+ *  
+ *  Returns an integer identifier for <i>obj</i>. The same number will
+ *  be returned on all calls to <code>id</code> for a given object, and
+ *  no two active objects will share an id.
+ *  <code>Object#object_id</code> is a different concept from the
+ *  <code>:name</code> notation, which returns the symbol id of
+ *  <code>name</code>. Replaces the deprecated <code>Object#id</code>.
+ */
+
+/*
+ *  call-seq:
+ *     obj.hash    => fixnum
+ *  
+ *  Generates a <code>Fixnum</code> hash value for this object. This
+ *  function must have the property that <code>a.eql?(b)</code> implies
+ *  <code>a.hash == b.hash</code>. The hash value is used by class
+ *  <code>Hash</code>. Any hash value that exceeds the capacity of a
+ *  <code>Fixnum</code> will be truncated before being used.
+ */
+
+VALUE
+rb_obj_id(VALUE obj)
+{
+    /*
+     *                32-bit VALUE space
+     *          MSB ------------------------ LSB
+     *  false   00000000000000000000000000000000
+     *  true    00000000000000000000000000000010
+     *  nil     00000000000000000000000000000100
+     *  undef   00000000000000000000000000000110
+     *  symbol  ssssssssssssssssssssssss00001110
+     *  object  oooooooooooooooooooooooooooooo00        = 0 (mod sizeof(RVALUE))
+     *  fixnum  fffffffffffffffffffffffffffffff1
+     *
+     *                    object_id space
+     *                                       LSB
+     *  false   00000000000000000000000000000000
+     *  true    00000000000000000000000000000010
+     *  nil     00000000000000000000000000000100
+     *  undef   00000000000000000000000000000110
+     *  symbol   000SSSSSSSSSSSSSSSSSSSSSSSSSSS0        S...S % A = 4 (S...S = s...s * A + 4)
+     *  object   oooooooooooooooooooooooooooooo0        o...o % A = 0
+     *  fixnum  fffffffffffffffffffffffffffffff1        bignum if required
+     *
+     *  where A = sizeof(RVALUE)/4
+     *
+     *  sizeof(RVALUE) is
+     *  20 if 32-bit, double is 4-byte aligned
+     *  24 if 32-bit, double is 8-byte aligned
+     *  40 if 64-bit
+     */
+    if (TYPE(obj) == T_SYMBOL) {
+        return (SYM2ID(obj) * sizeof(RVALUE) + (4 << 2)) | FIXNUM_FLAG;
+    }
+    if (SPECIAL_CONST_P(obj)) {
+        return LONG2NUM((long)obj);
+    }
+    return (VALUE)((long)obj|FIXNUM_FLAG);
+}
+
+/*
+ *  The <code>GC</code> module provides an interface to Ruby's mark and
+ *  sweep garbage collection mechanism. Some of the underlying methods
+ *  are also available via the <code>ObjectSpace</code> module.
+ */
+
 void
 Init_GC()
 {
@@ -1673,6 +2040,10 @@ Init_GC()
 
     source_filenames = st_init_strtable();
 
-    nomem_error = rb_exc_new2(rb_eNoMemError, "failed to allocate memory");
     rb_global_variable(&nomem_error);
+    nomem_error = rb_exc_new2(rb_eNoMemError, "failed to allocate memory");
+
+    rb_define_method(rb_mKernel, "hash", rb_obj_id, 0);
+    rb_define_method(rb_mKernel, "__id__", rb_obj_id, 0);
+    rb_define_method(rb_mKernel, "object_id", rb_obj_id, 0);
 }
