@@ -1,59 +1,3 @@
-#= open-uri.rb
-#
-#open-uri.rb is easy-to-use wrapper for net/http and net/ftp.
-#
-#== Example
-#
-#It is possible to open http/ftp URL as usual a file:
-#
-#  open("http://www.ruby-lang.org/") {|f|
-#    f.each_line {|line| p line}
-#  }
-#
-#The opened file has several methods for meta information as follows since
-#it is extended by OpenURI::Meta.
-#
-#  open("http://www.ruby-lang.org/en") {|f|
-#    f.each_line {|line| p line}
-#    p f.base_uri         # <URI::HTTP:0x40e6ef2 URL:http://www.ruby-lang.org/en/>
-#    p f.content_type     # "text/html"
-#    p f.charset          # "iso-8859-1"
-#    p f.content_encoding # []
-#    p f.last_modified    # Thu Dec 05 02:45:02 UTC 2002
-#  }
-#
-#Additional header fields can be specified by an optional hash argument.
-#
-#  open("http://www.ruby-lang.org/en/",
-#    "User-Agent" => "Ruby/#{RUBY_VERSION}",
-#    "From" => "foo@bar.invalid",
-#    "Referer" => "http://www.ruby-lang.org/") {|f|
-#    ...
-#  }
-#
-#The environment variables such as http_proxy and ftp_proxy are in effect by
-#default.  :proxy => nil disables proxy.
-#
-#  open("http://www.ruby-lang.org/en/raa.html",
-#    :proxy => nil) {|f|
-#    ...
-#  }
-#
-#URI objects can be opened in similar way.
-#
-#  uri = URI.parse("http://www.ruby-lang.org/en/")
-#  uri.open {|f|
-#    ...
-#  }
-#
-#URI objects can be read directly.
-#The returned string is also extended by OpenURI::Meta.
-#
-#  str = uri.read
-#  p str.base_uri
-#
-#Author:: Tanaka Akira <akr@m17n.org>
-
 require 'uri'
 require 'stringio'
 require 'time'
@@ -72,10 +16,11 @@ module Kernel
   #
   # Otherwise original open is called.
   #
-  # Since open-uri.rb provides URI::HTTP#open and URI::FTP#open,
+  # Since open-uri.rb provides URI::HTTP#open, URI::HTTPS#open and
+  # URI::FTP#open,
   # Kernel[#.]open can accepts such URIs and strings which begins with
-  # http:// and ftp://.  In this http and ftp case, the opened file object
-  # is extended by OpenURI::Meta.
+  # http://, https:// and ftp://.
+  # In these case, the opened file object is extended by OpenURI::Meta.
   def open(name, *rest, &block) # :doc:
     if name.respond_to?(:open)
       name.open(*rest, &block)
@@ -90,14 +35,67 @@ module Kernel
   module_function :open
 end
 
+# OpenURI is an easy-to-use wrapper for net/http, net/https and net/ftp.
+#
+#== Example
+#
+# It is possible to open http/https/ftp URL as usual like opening a file:
+#
+#   open("http://www.ruby-lang.org/") {|f|
+#     f.each_line {|line| p line}
+#   }
+#
+# The opened file has several methods for meta information as follows since
+# it is extended by OpenURI::Meta.
+#
+#   open("http://www.ruby-lang.org/en") {|f|
+#     f.each_line {|line| p line}
+#     p f.base_uri         # <URI::HTTP:0x40e6ef2 URL:http://www.ruby-lang.org/en/>
+#     p f.content_type     # "text/html"
+#     p f.charset          # "iso-8859-1"
+#     p f.content_encoding # []
+#     p f.last_modified    # Thu Dec 05 02:45:02 UTC 2002
+#   }
+#
+# Additional header fields can be specified by an optional hash argument.
+#
+#   open("http://www.ruby-lang.org/en/",
+#     "User-Agent" => "Ruby/#{RUBY_VERSION}",
+#     "From" => "foo@bar.invalid",
+#     "Referer" => "http://www.ruby-lang.org/") {|f|
+#     # ...
+#   }
+#
+# The environment variables such as http_proxy, https_proxy and ftp_proxy
+# are in effect by default.  :proxy => nil disables proxy.
+#
+#   open("http://www.ruby-lang.org/en/raa.html", :proxy => nil) {|f|
+#     # ...
+#   }
+#
+# URI objects can be opened in a similar way.
+#
+#   uri = URI.parse("http://www.ruby-lang.org/en/")
+#   uri.open {|f|
+#     # ...
+#   }
+#
+# URI objects can be read directly. The returned string is also extended by
+# OpenURI::Meta.
+#
+#   str = uri.read
+#   p str.base_uri
+#
+# Author:: Tanaka Akira <akr@m17n.org>
+
 module OpenURI
   Options = {
     :proxy => true,
     :progress_proc => true,
     :content_length_proc => true,
+    :http_basic_authentication => true,
   }
 
-
   def OpenURI.check_options(options) # :nodoc:
     options.each {|k, v|
       next unless Symbol === k
@@ -163,11 +161,7 @@ module OpenURI
     while true
       redirect = catch(:open_uri_redirect) {
         buf = Buffer.new
-        if proxy_uri = find_proxy.call(uri)
-          proxy_uri.proxy_open(buf, uri, options)
-        else
-          uri.direct_open(buf, options)
-        end
+        uri.buffer_open(buf, find_proxy.call(uri), options)
         nil
       }
       if redirect
@@ -179,6 +173,11 @@ module OpenURI
         unless OpenURI.redirectable?(uri, redirect)
           raise "redirection forbidden: #{uri} -> #{redirect}"
         end
+        if options.include? :http_basic_authentication
+          # send authentication only for the URI directly specified.
+          options = options.dup
+          options.delete :http_basic_authentication
+        end
         uri = redirect
         raise "HTTP redirection loop: #{uri}" if uri_set.include? uri.to_s
         uri_set[uri.to_s] = true
@@ -199,6 +198,96 @@ module OpenURI
     (/\A(?:http|ftp)\z/i =~ uri1.scheme && /\A(?:http|ftp)\z/i =~ uri2.scheme)
   end
 
+  def OpenURI.open_http(buf, target, proxy, options) # :nodoc:
+    if proxy
+      raise "Non-HTTP proxy URI: #{proxy}" if proxy.class != URI::HTTP
+    end
+
+    if target.userinfo && "1.9.0" <= RUBY_VERSION
+      # don't raise for 1.8 because compatibility.
+      raise ArgumentError, "userinfo not supported.  [RFC3986]"
+    end
+
+    require 'net/http'
+    klass = Net::HTTP
+    if URI::HTTP === target
+      # HTTP or HTTPS
+      if proxy
+        klass = Net::HTTP::Proxy(proxy.host, proxy.port)
+      end
+      target_host = target.host
+      target_port = target.port
+      request_uri = target.request_uri
+    else
+      # FTP over HTTP proxy
+      target_host = proxy.host
+      target_port = proxy.port
+      request_uri = target.to_s
+    end
+
+    http = klass.new(target_host, target_port)
+    if target.class == URI::HTTPS
+      require 'net/https'
+      http.use_ssl = true
+      http.verify_mode = OpenSSL::SSL::VERIFY_PEER
+      store = OpenSSL::X509::Store.new
+      store.set_default_paths
+      http.cert_store = store
+    end
+
+    header = {}
+    options.each {|k, v| header[k] = v if String === k }
+
+    resp = nil
+    http.start {
+      if target.class == URI::HTTPS
+        # xxx: information hiding violation
+        sock = http.instance_variable_get(:@socket)
+        if sock.respond_to?(:io)
+          sock = sock.io # 1.9
+        else
+          sock = sock.instance_variable_get(:@socket) # 1.8
+        end
+        sock.post_connection_check(target_host)
+      end
+      req = Net::HTTP::Get.new(request_uri, header)
+      if options.include? :http_basic_authentication
+        user, pass = options[:http_basic_authentication]
+        req.basic_auth user, pass
+      end
+      http.request(req) {|response|
+        resp = response
+        if options[:content_length_proc] && Net::HTTPSuccess === resp
+          if resp.key?('Content-Length')
+            options[:content_length_proc].call(resp['Content-Length'].to_i)
+          else
+            options[:content_length_proc].call(nil)
+          end
+        end
+        resp.read_body {|str|
+          buf << str
+          if options[:progress_proc] && Net::HTTPSuccess === resp
+            options[:progress_proc].call(buf.size)
+          end
+        }
+      }
+    }
+    io = buf.io
+    io.rewind
+    io.status = [resp.code, resp.message]
+    resp.each {|name,value| buf.io.meta_add_field name, value }
+    case resp
+    when Net::HTTPSuccess
+    when Net::HTTPMovedPermanently, # 301
+         Net::HTTPFound, # 302
+         Net::HTTPSeeOther, # 303
+         Net::HTTPTemporaryRedirect # 307
+      throw :open_uri_redirect, URI.parse(resp['location'])
+    else
+      raise OpenURI::HTTPError.new(io.status.join(' '), io)
+    end
+  end
+
   class HTTPError < StandardError
     def initialize(message, io)
       super(message)
@@ -222,14 +311,14 @@ module OpenURI
         require 'tempfile'
         io = Tempfile.new('open-uri')
         io.binmode
-        Meta.init io, @io if Meta === @io
+        Meta.init io, @io if @io.respond_to? :meta
         io << @io.string
         @io = io
       end
     end
 
     def io
-      Meta.init @io unless Meta === @io
+      Meta.init @io unless @io.respond_to? :meta
       @io
     end
   end
@@ -277,7 +366,7 @@ module OpenURI
 
     RE_LWS = /[\r\n\t ]+/n
     RE_TOKEN = %r{[^\x00- ()<>@,;:\\"/\[\]?={}\x7f]+}n
-    RE_QUOTED_STRING = %r{"(?:[\r\n\t !#-\[\]-~\x80-\xff]|\\[\x00-\x7f])"}n
+    RE_QUOTED_STRING = %r{"(?:[\r\n\t !#-\[\]-~\x80-\xff]|\\[\x00-\x7f])*"}n
     RE_PARAMETERS = %r{(?:;#{RE_LWS}?#{RE_TOKEN}#{RE_LWS}?=#{RE_LWS}?(?:#{RE_TOKEN}|#{RE_QUOTED_STRING})#{RE_LWS}?)*}n
 
     def content_type_parse # :nodoc:
@@ -362,60 +451,75 @@ module OpenURI
     # field for HTTP.
     # I.e. it is ignored for FTP without HTTP proxy.
     #
-    # The hash may include other option which key is a symbol:
+    # The hash may include other options which key is a symbol:
     #
-    # :proxy => "http://proxy.foo.com:8000/"
-    # :proxy => URI.parse("http://proxy.foo.com:8000/")
-    # :proxy => true
-    # :proxy => false
-    # :proxy => nil
+    # [:proxy]
+    #  Synopsis:
+    #    :proxy => "http://proxy.foo.com:8000/"
+    #    :proxy => URI.parse("http://proxy.foo.com:8000/")
+    #    :proxy => true
+    #    :proxy => false
+    #    :proxy => nil
+    #   
+    #  If :proxy option is specified, the value should be String, URI,
+    #  boolean or nil.
+    #  When String or URI is given, it is treated as proxy URI.
+    #  When true is given or the option itself is not specified,
+    #  environment variable `scheme_proxy' is examined.
+    #  `scheme' is replaced by `http', `https' or `ftp'.
+    #  When false or nil is given, the environment variables are ignored and
+    #  connection will be made to a server directly.
     #
-    #    If :proxy option is specified, the value should be String, URI,
-    #    boolean or nil.
-    #    When String or URI is given, it is treated as proxy URI.
-    #    When true is given or the option itself is not specified,
-    #    environment variable `scheme_proxy'(or `SCHEME_PROXY') is examined.
-    #    `scheme' is replaced by `http' or `ftp'.
-    #    When false or nil is given, the environment variables are ignored and
-    #    connection will be made to a server directly.
+    # [:http_basic_authentication]
+    #  Synopsis:
+    #    :http_basic_authentication=>[user, password]
     #
-    # :content_length_proc => lambda {|content_length| ... }
+    #  If :http_basic_authentication is specified,
+    #  the value should be an array which contains 2 strings:
+    #  username and password.
+    #  It is used for HTTP Basic authentication defined by RFC 2617.
     #
-    #   If :content_length_proc option is specified, the option value procedure
-    #   is called before actual transfer is started.
-    #   It takes one argument which is expected content length in bytes.
+    # [:content_length_proc]
+    #  Synopsis:
+    #    :content_length_proc => lambda {|content_length| ... }
+    # 
+    #  If :content_length_proc option is specified, the option value procedure
+    #  is called before actual transfer is started.
+    #  It takes one argument which is expected content length in bytes.
+    # 
+    #  If two or more transfer is done by HTTP redirection, the procedure
+    #  is called only one for a last transfer.
+    # 
+    #  When expected content length is unknown, the procedure is called with
+    #  nil.
+    #  It is happen when HTTP response has no Content-Length header.
     #
-    #   If two or more transfer is done by HTTP redirection, the procedure
-    #   is called only one for a last transfer.
+    # [:progress_proc]
+    #  Synopsis:
+    #    :progress_proc => lambda {|size| ...}
     #
-    #   When expected content length is unknown, the procedure is called with
-    #   nil.
-    #   It is happen when HTTP response has no Content-Length header.
+    #  If :progress_proc option is specified, the proc is called with one
+    #  argument each time when `open' gets content fragment from network.
+    #  The argument `size' `size' is a accumulated transfered size in bytes.
     #
-    # :progress_proc => lambda {|size| ...}
+    #  If two or more transfer is done by HTTP redirection, the procedure
+    #  is called only one for a last transfer.
     #
-    #   If :progress_proc option is specified, the proc is called with one
-    #   argument each time when `open' gets content fragment from network.
-    #   The argument `size' `size' is a accumulated transfered size in bytes.
+    #  :progress_proc and :content_length_proc are intended to be used for
+    #  progress bar.
+    #  For example, it can be implemented as follows using Ruby/ProgressBar.
     #
-    #   If two or more transfer is done by HTTP redirection, the procedure
-    #   is called only one for a last transfer.
-    #
-    #   :progress_proc and :content_length_proc are intended to be used for
-    #   progress bar.
-    #   For example, it can be implemented as follows using Ruby/ProgressBar.
-    #
-    #     pbar = nil
-    #     open("http://...",
-    #       :content_length_proc => lambda {|t|
-    #         if t && 0 < t
-    #           pbar = ProgressBar.new("...", t)
-    #           pbar.file_transfer_mode
-    #         end
-    #       },
-    #       :progress_proc => lambda {|s|
-    #         pbar.set s if pbar
-    #       }) {|f| ... }
+    #    pbar = nil
+    #    open("http://...",
+    #      :content_length_proc => lambda {|t|
+    #        if t && 0 < t
+    #          pbar = ProgressBar.new("...", t)
+    #          pbar.file_transfer_mode
+    #        end
+    #      },
+    #      :progress_proc => lambda {|s|
+    #        pbar.set s if pbar
+    #      }) {|f| ... }
     #
     # OpenURI::OpenRead#open returns an IO like object if block is not given.
     # Otherwise it yields the IO object and return the value of the block.
@@ -500,9 +604,6 @@ module URI
 
       if proxy_uri
         proxy_uri = URI.parse(proxy_uri)
-        unless URI::HTTP === proxy_uri
-          raise "Non-HTTP proxy URI: #{proxy_uri}"
-        end
         name = 'no_proxy'
         if no_proxy = ENV[name] || ENV[name.upcase]
           no_proxy.scan(/([^:,]*)(?::(\d+))?/) {|host, port|
@@ -521,83 +622,60 @@ module URI
   end
 
   class HTTP
-    def direct_open(buf, options) # :nodoc:
-      proxy_open(buf, request_uri, options)
-    end
-
-    def proxy_open(buf, uri, options) # :nodoc:
-      header = {}
-      options.each {|k, v| header[k] = v if String === k }
-
-      if uri.respond_to? :host
-        # According to RFC2616 14.23, Host: request-header field should be set
-        # an origin server.
-        # But net/http wrongly set a proxy server if an absolute URI is
-        # specified as a request URI.
-        # So open-uri override it here explicitly.
-        header['host'] = uri.host
-        header['host'] += ":#{uri.port}" if uri.port
-      end
-
-      require 'net/http'
-      resp = nil
-      Net::HTTP.start(self.host, self.port) {|http|
-        http.request_get(uri.to_s, header) {|response|
-          resp = response
-          if options[:content_length_proc] && Net::HTTPSuccess === resp
-            if resp.key?('Content-Length')
-              options[:content_length_proc].call(resp['Content-Length'].to_i)
-            else
-              options[:content_length_proc].call(nil)
-            end
-          end
-          resp.read_body {|str|
-            buf << str
-            if options[:progress_proc] && Net::HTTPSuccess === resp
-              options[:progress_proc].call(buf.size)
-            end
-          }
-        }
-      }
-      io = buf.io
-      io.rewind
-      io.status = [resp.code, resp.message]
-      resp.each {|name,value| buf.io.meta_add_field name, value }
-      case resp
-      when Net::HTTPSuccess
-      when Net::HTTPMovedPermanently, # 301
-           Net::HTTPFound, # 302
-           Net::HTTPSeeOther, # 303
-           Net::HTTPTemporaryRedirect # 307
-        throw :open_uri_redirect, URI.parse(resp['location'])
-      else
-        raise OpenURI::HTTPError.new(io.status.join(' '), io)
-      end
+    def buffer_open(buf, proxy, options) # :nodoc:
+      OpenURI.open_http(buf, self, proxy, options)
     end
 
     include OpenURI::OpenRead
   end
 
-  class HTTPS
-    def proxy_open(buf, uri, options) # :nodoc:
-      raise ArgumentError, "open-uri doesn't support https."
-    end
-  end
-
   class FTP
-    def direct_open(buf, options) # :nodoc:
+    def buffer_open(buf, proxy, options) # :nodoc:
+      if proxy
+        OpenURI.open_http(buf, self, proxy, options)
+        return
+      end
       require 'net/ftp'
+
+      directories = self.path.split(%r{/}, -1)
+      directories.shift if directories[0] == '' # strip a field before leading slash
+      directories.each {|d|
+        d.gsub!(/%([0-9A-Fa-f][0-9A-Fa-f])/) { [$1].pack("H2") }
+      }
+      unless filename = directories.pop
+        raise ArgumentError, "no filename: #{self.inspect}"
+      end
+      directories.each {|d|
+        if /[\r\n]/ =~ d
+          raise ArgumentError, "invalid directory: #{d.inspect}"
+        end
+      }
+      if /[\r\n]/ =~ filename
+        raise ArgumentError, "invalid filename: #{filename.inspect}"
+      end
+      typecode = self.typecode
+      if typecode && /\A[aid]\z/ !~ typecode
+        raise ArgumentError, "invalid typecode: #{typecode.inspect}"
+      end
+
+      # The access sequence is defined by RFC 1738
+      ftp = Net::FTP.open(self.host)
       # todo: extract user/passwd from .netrc.
       user = 'anonymous'
       passwd = nil
       user, passwd = self.userinfo.split(/:/) if self.userinfo
-
-      ftp = Net::FTP.open(self.host)
       ftp.login(user, passwd)
-      if options[:content_length_proc]
-        options[:content_length_proc].call(ftp.size(self.path))
+      directories.each {|cwd|
+        ftp.voidcmd("CWD #{cwd}")
+      }
+      if typecode
+        # xxx: typecode D is not handled.
+        ftp.voidcmd("TYPE #{typecode.upcase}")
       end
-      ftp.getbinaryfile(self.path, '/dev/null', Net::FTP::DEFAULT_BLOCKSIZE) {|str|
+      if options[:content_length_proc]
+        options[:content_length_proc].call(ftp.size(filename))
+      end
+      ftp.retrbinary("RETR #{filename}", 4096) { |str|
         buf << str
         options[:progress_proc].call(buf.size) if options[:progress_proc]
       }
