@@ -2,8 +2,8 @@
 
   marshal.c -
 
-  $Author: nobu $
-  $Date: 2003/12/22 08:23:54 $
+  $Author: akr $
+  $Date: 2005/12/14 03:04:14 $
   created at: Thu Apr 27 16:30:01 JST 1995
 
   Copyright (C) 1993-2003 Yukihiro Matsumoto
@@ -98,6 +98,24 @@ struct dump_call_arg {
     int limit;
 };
 
+static VALUE
+class2path(klass)
+    VALUE klass;
+{
+    VALUE path = rb_class_path(klass);
+    char *n = RSTRING(path)->ptr;
+
+    if (n[0] == '#') {
+	rb_raise(rb_eTypeError, "can't dump anonymous %s %s",
+		 (TYPE(klass) == T_CLASS ? "class" : "module"),
+		 n);
+    }
+    if (rb_path2class(n) != rb_class_real(klass)) {
+	rb_raise(rb_eTypeError, "%s can't be referred", n);
+    }
+    return path;
+}
+
 static void w_long _((long, struct dump_arg*));
 
 static void
@@ -382,7 +400,7 @@ w_class(type, obj, arg, check)
     VALUE klass = CLASS_OF(obj);
     w_extended(klass, arg, check);
     w_byte(type, arg);
-    path = rb_class2name(klass);
+    path = RSTRING(class2path(rb_class_real(klass)))->ptr;
     w_unique(path, arg);
 }
 
@@ -397,7 +415,7 @@ w_uclass(obj, base_klass, arg)
     klass = rb_class_real(klass);
     if (klass != base_klass) {
 	w_byte(TYPE_UCLASS, arg);
-	w_unique(rb_class2name(klass), arg);
+	w_unique(RSTRING(class2path(klass))->ptr, arg);
     }
 }
 
@@ -419,7 +437,7 @@ w_ivar(tbl, arg)
 {
     if (tbl) {
 	w_long(tbl->num_entries, arg->arg);
-	st_foreach(tbl, w_obj_each, (st_data_t)arg);
+	st_foreach_safe(tbl, w_obj_each, (st_data_t)arg);
     }
     else {
 	w_long(0, arg->arg);
@@ -517,11 +535,7 @@ w_object(obj, arg, limit)
 	    }
 	    w_byte(TYPE_CLASS, arg);
 	    {
-		VALUE path = rb_class_path(obj);
-		if (RSTRING(path)->ptr[0] == '#') {
-		    rb_raise(rb_eTypeError, "can't dump anonymous class %s",
-			     RSTRING(path)->ptr);
-		}
+		VALUE path = class2path(obj);
 		w_bytes(RSTRING(path)->ptr, RSTRING(path)->len, arg);
 	    }
 	    break;
@@ -529,11 +543,7 @@ w_object(obj, arg, limit)
 	  case T_MODULE:
 	    w_byte(TYPE_MODULE, arg);
 	    {
-		VALUE path = rb_class_path(obj);
-		if (RSTRING(path)->ptr[0] == '#') {
-		    rb_raise(rb_eTypeError, "can't dump anonymous module %s",
-			     RSTRING(path)->ptr);
-		}
+		VALUE path = class2path(obj);
 		w_bytes(RSTRING(path)->ptr, RSTRING(path)->len, arg);
 	    }
 	    break;
@@ -605,13 +615,13 @@ w_object(obj, arg, limit)
 	    }
 	    else if (FL_TEST(obj, FL_USER2)) {
 		/* FL_USER2 means HASH_PROC_DEFAULT (see hash.c) */
-		rb_raise(rb_eTypeError, "cannot dump hash with default proc");
+		rb_raise(rb_eTypeError, "can't dump hash with default proc");
 	    }
 	    else {
 		w_byte(TYPE_HASH_DEF, arg);
 	    }
 	    w_long(RHASH(obj)->tbl->num_entries, arg);
-	    st_foreach(RHASH(obj)->tbl, hash_each, (st_data_t)&c_arg);
+	    rb_hash_foreach(obj, hash_each, (st_data_t)&c_arg);
 	    if (!NIL_P(RHASH(obj)->ifnone)) {
 		w_object(RHASH(obj)->ifnone, arg, limit);
 	    }
@@ -625,10 +635,7 @@ w_object(obj, arg, limit)
 		long i;
 
 		w_long(len, arg);
-		mem = rb_struct_iv_get(rb_obj_class(obj), "__member__");
-		if (mem == Qnil) {
-		    rb_raise(rb_eTypeError, "uninitialized struct");
-		}
+		mem = rb_struct_members(obj);
 		for (i=0; i<len; i++) {
 		    w_symbol(SYM2ID(RARRAY(mem)->ptr[i]), arg);
 		    w_object(RSTRUCT(obj)->ptr[i], arg, limit);
@@ -691,6 +698,32 @@ dump_ensure(arg)
     return 0;
 }
 
+/*
+ * call-seq:
+ *      dump( obj [, anIO] , limit=--1 ) => anIO
+ *
+ * Serializes obj and all descendent objects. If anIO is
+ * specified, the serialized data will be written to it, otherwise the
+ * data will be returned as a String. If limit is specified, the
+ * traversal of subobjects will be limited to that depth. If limit is
+ * negative, no checking of depth will be performed.
+ *
+ *     class Klass
+ *       def initialize(str)
+ *         @str = str
+ *       end
+ *       def sayHello
+ *         @str
+ *       end
+ *     end
+ *
+ * (produces no output)
+ *
+ *     o = Klass.new("hello\n")
+ *     data = Marshal.dump(o)
+ *     obj = Marshal.load(data)
+ *     obj.sayHello   #=> "hello\n"
+ */
 static VALUE
 marshal_dump(argc, argv)
     int argc;
@@ -746,7 +779,8 @@ marshal_dump(argc, argv)
 }
 
 struct load_arg {
-    char *ptr, *end;
+    VALUE src;
+    long offset;
     st_table *symbols;
     VALUE data;
     VALUE proc;
@@ -761,18 +795,20 @@ r_byte(arg)
 {
     int c;
 
-    if (!arg->end) {
-	VALUE src = (VALUE)arg->ptr;
+    if (TYPE(arg->src) == T_STRING) {
+	if (RSTRING(arg->src)->len > arg->offset) {
+	    c = (unsigned char)RSTRING(arg->src)->ptr[arg->offset++];
+	}
+	else {
+	    rb_raise(rb_eArgError, "marshal data too short");
+	}
+    }
+    else {
+	VALUE src = arg->src;
 	VALUE v = rb_funcall2(src, s_getc, 0, 0);
 	if (NIL_P(v)) rb_eof_error();
 	c = (unsigned char)FIX2INT(v);
     }
-    else if (arg->ptr < arg->end) {
-	c = *(unsigned char*)arg->ptr++;
-    }
-    else {
-	rb_raise(rb_eArgError, "marshal data too short");
-    }
     return c;
 }
 
@@ -835,8 +871,19 @@ r_bytes0(len, arg)
 {
     VALUE str;
 
-    if (!arg->end) {
-	VALUE src = (VALUE)arg->ptr;
+    if (len == 0) return rb_str_new(0, 0);
+    if (TYPE(arg->src) == T_STRING) {
+	if (RSTRING(arg->src)->len > arg->offset) {
+	    str = rb_str_new(RSTRING(arg->src)->ptr+arg->offset, len);
+	    arg->offset += len;
+	}
+	else {
+	  too_short:
+	    rb_raise(rb_eArgError, "marshal data too short");
+	}
+    }
+    else {
+	VALUE src = arg->src;
 	VALUE n = LONG2NUM(len);
 	str = rb_funcall2(src, s_read, 1, &n);
 	if (NIL_P(str)) goto too_short;
@@ -844,14 +891,6 @@ r_bytes0(len, arg)
 	if (RSTRING(str)->len != len) goto too_short;
 	if (OBJ_TAINTED(str)) arg->taint = Qtrue;
     }
-    else {
-	if (arg->ptr + len > arg->end) {
-	  too_short:
-	    rb_raise(rb_eArgError, "marshal data too short");
-	}
-	str = rb_str_new(arg->ptr, len);
-	arg->ptr += len;
-    }
     return str;
 }
 
@@ -1011,7 +1050,7 @@ r_object0(arg, proc, ivp, extmod)
 	      format_error:
 		rb_raise(rb_eArgError, "dump format error (user class)");
 	    }
-	    if (TYPE(v) == T_MODULE || !RTEST(rb_funcall(c, '<', 1, RBASIC(v)->klass))) {
+	    if (TYPE(v) == T_MODULE || !RTEST(rb_class_inherited_p(c, RBASIC(v)->klass))) {
 		VALUE tmp = rb_obj_alloc(c);
 
 		if (TYPE(v) != TYPE(tmp)) goto format_error;
@@ -1068,7 +1107,7 @@ r_object0(arg, proc, ivp, extmod)
 	{
 	    long len;
 	    BDIGIT *digits;
-	    VALUE data;
+	    volatile VALUE data;
 
 	    NEWOBJ(big, struct RBignum);
 	    OBJSETUP(big, rb_cBignum, T_BIGNUM);
@@ -1159,7 +1198,7 @@ r_object0(arg, proc, ivp, extmod)
 	    ID slot;
 
 	    klass = path2class(r_unique(arg));
-	    mem = rb_struct_iv_get(klass, "__member__");
+	    mem = rb_struct_s_members(klass);
 	    if (mem == Qnil) {
 		rb_raise(rb_eTypeError, "uninitialized struct");
 	    }
@@ -1332,6 +1371,17 @@ load_ensure(arg)
     return 0;
 }
 
+/*
+ * call-seq:
+ *     load( source [, proc] ) => obj
+ *     restore( source [, proc] ) => obj
+ * 
+ * Returns the result of converting the serialized data in source into a
+ * Ruby object (possibly with associated subordinate objects). source
+ * may be either an instance of IO or an object that responds to
+ * to_str. If proc is specified, it will be passed each object as it
+ * is deserialized.
+ */
 static VALUE
 marshal_load(argc, argv)
     int argc;
@@ -1346,20 +1396,18 @@ marshal_load(argc, argv)
     if (rb_respond_to(port, rb_intern("to_str"))) {
 	arg.taint = OBJ_TAINTED(port); /* original taintedness */
 	StringValue(port);	       /* possible conversion */
-	arg.ptr = RSTRING(port)->ptr;
-	arg.end = arg.ptr + RSTRING(port)->len;
     }
     else if (rb_respond_to(port, s_getc) && rb_respond_to(port, s_read)) {
 	if (rb_respond_to(port, s_binmode)) {
 	    rb_funcall2(port, s_binmode, 0, 0);
 	}
 	arg.taint = Qtrue;
-	arg.ptr = (char *)port;
-	arg.end = 0;
     }
     else {
 	rb_raise(rb_eTypeError, "instance of IO needed");
     }
+    arg.src = port;
+    arg.offset = 0;
 
     major = r_byte(&arg);
     minor = r_byte(&arg);
@@ -1383,6 +1431,39 @@ marshal_load(argc, argv)
     return v;
 }
 
+/*
+ * The marshaling library converts collections of Ruby objects into a
+ * byte stream, allowing them to be stored outside the currently
+ * active script. This data may subsequently be read and the original
+ * objects reconstituted.
+ * Marshaled data has major and minor version numbers stored along
+ * with the object information. In normal use, marshaling can only
+ * load data written with the same major version number and an equal
+ * or lower minor version number. If Ruby's ``verbose'' flag is set
+ * (normally using -d, -v, -w, or --verbose) the major and minor
+ * numbers must match exactly. Marshal versioning is independent of
+ * Ruby's version numbers. You can extract the version by reading the
+ * first two bytes of marshaled data.
+ *
+ *     str = Marshal.dump("thing")
+ *     RUBY_VERSION   #=> "1.8.0"
+ *     str[0]         #=> 4
+ *     str[1]         #=> 8
+ *
+ * Some objects cannot be dumped: if the objects to be dumped include
+ * bindings, procedure or method objects, instances of class IO, or
+ * singleton objects, a TypeError will be raised.
+ * If your class has special serialization needs (for example, if you
+ * want to serialize in some specific format), or if it contains
+ * objects that would otherwise not be serializable, you can implement
+ * your own serialization strategy by defining two methods, _dump and
+ * _load:
+ * The instance method _dump should return a String object containing
+ * all the information necessary to reconstitute objects of this class
+ * and all referenced objects up to a maximum depth given as an integer
+ * parameter (a value of -1 implies that you should disable depth checking).
+ * The class method _load should take a String and return an object of this class.
+ */
 void
 Init_marshal()
 {
