@@ -2,8 +2,8 @@
 
   numeric.c -
 
-  $Author: dave $
-  $Date: 2003/12/23 16:21:16 $
+  $Author: matz $
+  $Date: 2006/05/01 03:46:46 $
   created at: Fri Aug 13 18:33:09 JST 1993
 
   Copyright (C) 1993-2003 Yukihiro Matsumoto
@@ -12,6 +12,7 @@
 
 #include "ruby.h"
 #include "env.h"
+#include <ctype.h>
 #include <math.h>
 #include <stdio.h>
 
@@ -78,6 +79,23 @@ rb_num_zerodiv()
     rb_raise(rb_eZeroDivError, "divided by 0");
 }
 
+
+/*
+ *  call-seq:
+ *     num.coerce(numeric)   => array
+ *  
+ *  If <i>aNumeric</i> is the same type as <i>num</i>, returns an array
+ *  containing <i>aNumeric</i> and <i>num</i>. Otherwise, returns an
+ *  array with both <i>aNumeric</i> and <i>num</i> represented as
+ *  <code>Float</code> objects. This coercion mechanism is used by
+ *  Ruby to handle mixed-type numeric operations: it is intended to
+ *  find a compatible common type between the two operands of the operator.
+ *     
+ *     1.coerce(2.5)   #=> [2.5, 1.0]
+ *     1.2.coerce(3)   #=> [3.0, 1.2]
+ *     1.coerce(2)     #=> [2, 1]
+ */
+
 static VALUE
 num_coerce(x, y)
     VALUE x, y;
@@ -162,6 +180,11 @@ rb_num_coerce_relop(x, y)
     return c;
 }
 
+/*
+ * Trap attempts to add methods to <code>Numeric</code> objects. Always
+ * raises a <code>TypeError</code>
+ */
+
 static VALUE
 num_sadded(x, name)
     VALUE x, name;
@@ -175,6 +198,7 @@ num_sadded(x, name)
     return Qnil;		/* not reached */
 }
 
+/* :nodoc: */
 static VALUE
 num_init_copy(x, y)
     VALUE x, y;
@@ -184,6 +208,13 @@ num_init_copy(x, y)
     return Qnil;		/* not reached */
 }
 
+/*
+ *  call-seq:
+ *     +num    => num
+ *  
+ *  Unary Plus---Returns the receiver's value.
+ */
+
 static VALUE
 num_uplus(num)
     VALUE num;
@@ -191,6 +222,13 @@ num_uplus(num)
     return num;
 }
 
+/*
+ *  call-seq:
+ *     -num    => numeric
+ *  
+ *  Unary Minus---Returns the receiver's value, negated.
+ */
+
 static VALUE
 num_uminus(num)
     VALUE num;
@@ -203,6 +241,13 @@ num_uminus(num)
     return rb_funcall(zero, '-', 1, num);
 }
 
+/*
+ *  call-seq:
+ *     num.quo(numeric)    =>   result
+ *  
+ *  Equivalent to <code>Numeric#/</code>, but overridden in subclasses.
+ */
+
 static VALUE
 num_quo(x, y)
     VALUE x, y;
@@ -210,13 +255,67 @@ num_quo(x, y)
     return rb_funcall(x, '/', 1, y);
 }
 
+
+static VALUE num_floor(VALUE num);
+
+/*
+ *  call-seq:
+ *     num.div(numeric)    => integer
+ *  
+ *  Uses <code>/</code> to perform division, then converts the result to
+ *  an integer. <code>Numeric</code> does not define the <code>/</code>
+ *  operator; this is left to subclasses.
+ */
+
 static VALUE
 num_div(x, y)
     VALUE x, y;
 {
-    return rb_Integer(rb_funcall(x, '/', 1, y));
+    return num_floor(rb_funcall(x, '/', 1, y));
 }
 
+
+
+/*
+ *  call-seq:
+ *     num.divmod( aNumeric ) -> anArray
+ *  
+ *  Returns an array containing the quotient and modulus obtained by
+ *  dividing <i>num</i> by <i>aNumeric</i>. If <code>q, r =
+ *  x.divmod(y)</code>, then
+ *
+ *      q = floor(float(x)/float(y))
+ *      x = q*y + r
+ *     
+ *  The quotient is rounded toward -infinity, as shown in the following table:
+ *     
+ *     a    |  b  |  a.divmod(b)  |   a/b   | a.modulo(b) | a.remainder(b)
+ *    ------+-----+---------------+---------+-------------+---------------
+ *     13   |  4  |   3,    1     |   3     |    1        |     1
+ *    ------+-----+---------------+---------+-------------+---------------
+ *     13   | -4  |  -4,   -3     |  -3     |   -3        |     1
+ *    ------+-----+---------------+---------+-------------+---------------
+ *    -13   |  4  |  -4,    3     |  -4     |    3        |    -1
+ *    ------+-----+---------------+---------+-------------+---------------
+ *    -13   | -4  |   3,   -1     |   3     |   -1        |    -1
+ *    ------+-----+---------------+---------+-------------+---------------
+ *     11.5 |  4  |   2,    3.5   |   2.875 |    3.5      |     3.5
+ *    ------+-----+---------------+---------+-------------+---------------
+ *     11.5 | -4  |  -3,   -0.5   |  -2.875 |   -0.5      |     3.5
+ *    ------+-----+---------------+---------+-------------+---------------
+ *    -11.5 |  4  |  -3,    0.5   |  -2.875 |    0.5      |    -3.5
+ *    ------+-----+---------------+---------+-------------+---------------
+ *    -11.5 | -4  |   2    -3.5   |   2.875 |   -3.5      |    -3.5
+ *
+ *
+ *  Examples
+ *     11.divmod(3)         #=> [3, 2]
+ *     11.divmod(-3)        #=> [-4, -1]
+ *     11.divmod(3.5)       #=> [3, 0.5]
+ *     (-11).divmod(3.5)    #=> [-4, 3.0]
+ *     (11.5).divmod(3.5)   #=> [3, 1.0]
+ */
+
 static VALUE
 num_divmod(x, y)
     VALUE x, y;
@@ -224,6 +323,14 @@ num_divmod(x, y)
     return rb_assoc_new(num_div(x, y), rb_funcall(x, '%', 1, y));
 }
 
+/*
+ *  call-seq:
+ *     num.modulo(numeric)    => result
+ *  
+ *  Equivalent to
+ *  <i>num</i>.<code>divmod(</code><i>aNumeric</i><code>)[1]</code>.
+ */
+
 static VALUE
 num_modulo(x, y)
     VALUE x, y;
@@ -231,6 +338,18 @@ num_modulo(x, y)
     return rb_funcall(x, '%', 1, y);
 }
 
+/*
+ *  call-seq:
+ *     num.remainder(numeric)    => result
+ *  
+ *  If <i>num</i> and <i>numeric</i> have different signs, returns
+ *  <em>mod</em>-<i>numeric</i>; otherwise, returns <em>mod</em>. In
+ *  both cases <em>mod</em> is the value
+ *  <i>num</i>.<code>modulo(</code><i>numeric</i><code>)</code>. The
+ *  differences between <code>remainder</code> and modulo
+ *  (<code>%</code>) are shown in the table under <code>Numeric#divmod</code>.
+ */
+
 static VALUE
 num_remainder(x, y)
     VALUE x, y;
@@ -247,6 +366,14 @@ num_remainder(x, y)
     return z;
 }
 
+/*
+ *  call-seq:
+ *     num.integer? -> true or false
+ *  
+ *  Returns <code>true</code> if <i>num</i> is an <code>Integer</code>
+ *  (including <code>Fixnum</code> and <code>Bignum</code>).
+ */
+
 static VALUE
 num_int_p(num)
     VALUE num;
@@ -254,6 +381,17 @@ num_int_p(num)
     return Qfalse;
 }
 
+/*
+ *  call-seq:
+ *     num.abs   => num or numeric
+ *  
+ *  Returns the absolute value of <i>num</i>.
+ *     
+ *     12.abs         #=> 12
+ *     (-34.56).abs   #=> 34.56
+ *     -34.56.abs     #=> 34.56
+ */
+
 static VALUE
 num_abs(num)
     VALUE num;
@@ -264,6 +402,14 @@ num_abs(num)
     return num;
 }
 
+
+/*
+ *  call-seq:
+ *     num.zero?    => true or false
+ *  
+ *  Returns <code>true</code> if <i>num</i> has a zero value.
+ */
+
 static VALUE
 num_zero_p(num)
     VALUE num;
@@ -274,6 +420,19 @@ num_zero_p(num)
     return Qfalse;
 }
 
+
+/*
+ *  call-seq:
+ *     num.nonzero?    => num or nil
+ *  
+ *  Returns <i>num</i> if <i>num</i> is not zero, <code>nil</code>
+ *  otherwise. This behavior is useful when chaining comparisons:
+ *     
+ *     a = %w( z Bb bB bb BB a aA Aa AA A )
+ *     b = a.sort {|a,b| (a.downcase <=> b.downcase).nonzero? || a <=> b }
+ *     b   #=> ["A", "a", "AA", "Aa", "aA", "BB", "Bb", "bB", "bb", "z"]
+ */
+
 static VALUE
 num_nonzero_p(num)
     VALUE num;
@@ -284,6 +443,14 @@ num_nonzero_p(num)
     return num;
 }
 
+/*
+ *  call-seq:
+ *     num.to_int    => integer
+ *  
+ *  Invokes the child class's <code>to_i</code> method to convert
+ *  <i>num</i> to an integer.
+ */
+
 static VALUE
 num_to_int(num)
     VALUE num;
@@ -326,37 +493,28 @@ flo_to_s(flt)
     VALUE flt;
 {
     char buf[32];
-    char *fmt = "%.15g";
     double value = RFLOAT(flt)->value;
-    double avalue, d1, d2;
+    char *p, *e;
 
     if (isinf(value))
 	return rb_str_new2(value < 0 ? "-Infinity" : "Infinity");
     else if(isnan(value))
 	return rb_str_new2("NaN");
-    
-    avalue = fabs(value);
-    if (avalue == 0.0) {
-	fmt = "%.1f";
-    }
-    else if (avalue < 1.0e-3) {
-	d1 = avalue;
-	while (d1 < 1.0) d1 *= 10.0;
-	d1 = modf(d1, &d2);
-	if (d1 == 0) fmt = "%.1e";
-    }    
-    else if (avalue >= 1.0e15) {
-	d1 = avalue;
-	while (d1 > 10.0) d1 /= 10.0;
-	d1 = modf(d1, &d2);
-	if (d1 == 0) fmt = "%.1e";
-	else fmt = "%.16e";
-    }    
-    else if ((d1 = modf(value, &d2)) == 0) {
-	fmt = "%.1f";
-    }
-    sprintf(buf, fmt, value);
 
+    sprintf(buf, "%#.15g", value); /* ensure to print decimal point */
+    if (!(e = strchr(buf, 'e'))) {
+	e = buf + strlen(buf);
+    }
+    if (!ISDIGIT(e[-1])) { /* reformat if ended with decimal point (ex 111111111111111.) */
+	sprintf(buf, "%#.14e", value);
+	if (!(e = strchr(buf, 'e'))) {
+	    e = buf + strlen(buf);
+	}
+    }
+    p = e;
+    while (p[-1]=='0' && ISDIGIT(p[-2]))
+	p--;
+    memmove(p, e, strlen(e)+1);
     return rb_str_new2(buf);
 }
 
@@ -437,7 +595,7 @@ flo_minus(x, y)
  * call-seq:
  *   float * other   => float
  *
- * Returns a new float with is the product of <code>float</code>
+ * Returns a new float which is the product of <code>float</code>
  * and <code>other</code>.
  */
 
@@ -519,7 +677,7 @@ flodivmod(x, y, divp, modp)
  *     flt % other         => float
  *     flt.modulo(other)   => float
  *  
- *  Return the modulo after dividion of <code>flt</code> by <code>other</code>.
+ *  Return the modulo after division of <code>flt</code> by <code>other</code>.
  *     
  *     6543.21.modulo(137)      #=> 104.21
  *     6543.21.modulo(137.24)   #=> 92.9299999999996
@@ -559,7 +717,8 @@ static VALUE
 flo_divmod(x, y)
     VALUE x, y;
 {
-    double fy, div, mod;
+    double fy, div, mod, val;
+    volatile VALUE a, b;
 
     switch (TYPE(y)) {
       case T_FIXNUM:
@@ -575,7 +734,15 @@ flo_divmod(x, y)
 	return rb_num_coerce_bin(x, y);
     }
     flodivmod(RFLOAT(x)->value, fy, &div, &mod);
-    return rb_assoc_new(rb_float_new(div), rb_float_new(mod));
+    if (FIXABLE(div)) {
+        val = div;
+        a = LONG2FIX(val);
+    }
+    else {
+        a = rb_dbl2big(div);
+    }
+    b = rb_float_new(mod);
+    return rb_assoc_new(a, b);
 }
 
 /*
@@ -602,6 +769,18 @@ flo_pow(x, y)
     }
 }
 
+/*
+ *  call-seq:
+ *     num.eql?(numeric)    => true or false
+ *  
+ *  Returns <code>true</code> if <i>num</i> and <i>numeric</i> are the
+ *  same type and have equal values.
+ *     
+ *     1 == 1.0          #=> true
+ *     1.eql?(1.0)       #=> false
+ *     (1.0).eql?(1.0)   #=> true
+ */
+
 static VALUE
 num_eql(x, y)
     VALUE x, y;
@@ -611,6 +790,14 @@ num_eql(x, y)
     return rb_equal(x, y);
 }
 
+/*
+ *  call-seq:
+ *     num <=> other -> 0 or nil
+ *  
+ *  Returns zero if <i>num</i> equals <i>other</i>, <code>nil</code>
+ *  otherwise.
+ */
+
 static VALUE
 num_cmp(x, y)
     VALUE x, y;
@@ -643,7 +830,7 @@ static VALUE
 flo_eq(x, y)
     VALUE x, y;
 {
-    double a, b;
+    volatile double a, b;
 
     switch (TYPE(y)) {
       case T_FIXNUM:
@@ -654,12 +841,13 @@ flo_eq(x, y)
 	break;
       case T_FLOAT:
 	b = RFLOAT(y)->value;
+	if (isnan(b)) return Qfalse;
 	break;
       default:
 	return num_equal(x, y);
     }
     a = RFLOAT(x)->value;
-    if (isnan(a) || isnan(b)) return Qfalse;
+    if (isnan(a)) return Qfalse;
     return (a == b)?Qtrue:Qfalse;
 }
 
@@ -759,12 +947,13 @@ flo_gt(x, y)
 
       case T_FLOAT:
 	b = RFLOAT(y)->value;
+	if (isnan(b)) return Qfalse;
 	break;
 
       default:
 	return rb_num_coerce_relop(x, y);
     }
-    if (isnan(a) || isnan(b)) return Qfalse;
+    if (isnan(a)) return Qfalse;
     return (a > b)?Qtrue:Qfalse;
 }
 
@@ -794,12 +983,13 @@ flo_ge(x, y)
 
       case T_FLOAT:
 	b = RFLOAT(y)->value;
+	if (isnan(b)) return Qfalse;
 	break;
 
       default:
 	return rb_num_coerce_relop(x, y);
     }
-    if (isnan(a) || isnan(b)) return Qfalse;
+    if (isnan(a)) return Qfalse;
     return (a >= b)?Qtrue:Qfalse;
 }
 
@@ -828,12 +1018,13 @@ flo_lt(x, y)
 
       case T_FLOAT:
 	b = RFLOAT(y)->value;
+	if (isnan(b)) return Qfalse;
 	break;
 
       default:
 	return rb_num_coerce_relop(x, y);
     }
-    if (isnan(a) || isnan(b)) return Qfalse;
+    if (isnan(a)) return Qfalse;
     return (a < b)?Qtrue:Qfalse;
 }
 
@@ -863,12 +1054,13 @@ flo_le(x, y)
 
       case T_FLOAT:
 	b = RFLOAT(y)->value;
+	if (isnan(b)) return Qfalse;
 	break;
 
       default:
 	return rb_num_coerce_relop(x, y);
     }
-    if (isnan(a) || isnan(b)) return Qfalse;
+    if (isnan(a)) return Qfalse;
     return (a <= b)?Qtrue:Qfalse;
 }
 
@@ -1135,6 +1327,19 @@ flo_truncate(num)
     return LONG2FIX(val);
 }
 
+
+/*
+ *  call-seq:
+ *     num.floor    => integer
+ *  
+ *  Returns the largest integer less than or equal to <i>num</i>.
+ *  <code>Numeric</code> implements this by converting <i>anInteger</i>
+ *  to a <code>Float</code> and invoking <code>Float#floor</code>.
+ *     
+ *     1.floor      #=> 1
+ *     (-1).floor   #=> -1
+ */
+
 static VALUE
 num_floor(num)
     VALUE num;
@@ -1142,6 +1347,22 @@ num_floor(num)
     return flo_floor(rb_Float(num));
 }
 
+
+/*
+ *  call-seq:
+ *     num.ceil    => integer
+ *  
+ *  Returns the smallest <code>Integer</code> greater than or equal to
+ *  <i>num</i>. Class <code>Numeric</code> achieves this by converting
+ *  itself to a <code>Float</code> then invoking
+ *  <code>Float#ceil</code>.
+ *     
+ *     1.ceil        #=> 1
+ *     1.2.ceil      #=> 2
+ *     (-1.2).ceil   #=> -1
+ *     (-1.0).ceil   #=> -1
+ */
+
 static VALUE
 num_ceil(num)
     VALUE num;
@@ -1149,6 +1370,15 @@ num_ceil(num)
     return flo_ceil(rb_Float(num));
 }
 
+/*
+ *  call-seq:
+ *     num.round    => integer
+ *  
+ *  Rounds <i>num</i> to the nearest integer. <code>Numeric</code>
+ *  implements this by converting itself to a
+ *  <code>Float</code> and invoking <code>Float#round</code>.
+ */
+
 static VALUE
 num_round(num)
     VALUE num;
@@ -1156,6 +1386,15 @@ num_round(num)
     return flo_round(rb_Float(num));
 }
 
+/*
+ *  call-seq:
+ *     num.truncate    => integer
+ *  
+ *  Returns <i>num</i> truncated to an integer. <code>Numeric</code>
+ *  implements this by converting its value to a float and invoking
+ *  <code>Float#truncate</code>.
+ */
+
 static VALUE
 num_truncate(num)
     VALUE num;
@@ -1163,6 +1402,34 @@ num_truncate(num)
     return flo_truncate(rb_Float(num));
 }
 
+
+/*
+ *  call-seq:
+ *     num.step(limit, step ) {|i| block }     => num
+ *  
+ *  Invokes <em>block</em> with the sequence of numbers starting at
+ *  <i>num</i>, incremented by <i>step</i> on each call. The loop
+ *  finishes when the value to be passed to the block is greater than
+ *  <i>limit</i> (if <i>step</i> is positive) or less than
+ *  <i>limit</i> (if <i>step</i> is negative). If all the arguments are
+ *  integers, the loop operates using an integer counter. If any of the
+ *  arguments are floating point numbers, all are converted to floats,
+ *  and the loop is executed <i>floor(n + n*epsilon)+ 1</i> times,
+ *  where <i>n = (limit - num)/step</i>. Otherwise, the loop
+ *  starts at <i>num</i>, uses either the <code><</code> or
+ *  <code>></code> operator to compare the counter against
+ *  <i>limit</i>, and increments itself using the <code>+</code>
+ *  operator.
+ *     
+ *     1.step(10, 2) { |i| print i, " " }
+ *     Math::E.step(Math::PI, 0.2) { |f| print f, " " }
+ *     
+ *  <em>produces:</em>
+ *     
+ *     1 3 5 7 9
+ *     2.71828182845905 2.91828182845905 3.11828182845905
+ */
+
 static VALUE
 num_step(argc, argv, from)
     int argc;
@@ -1184,7 +1451,7 @@ num_step(argc, argv, from)
 	    rb_raise(rb_eArgError, "wrong number of arguments");
 	}
 	if (rb_equal(step, INT2FIX(0))) {
-	    rb_raise(rb_eArgError, "step cannot be 0");
+	    rb_raise(rb_eArgError, "step can't be 0");
 	}
     }
 
@@ -1270,9 +1537,6 @@ rb_num2long(val)
       case T_BIGNUM:
 	return rb_big2long(val);
 
-      case T_SYMBOL:
-	rb_warning("treating Symbol as an integer");
-	/* fall through */
       default:
 	val = rb_to_int(val);
 	return NUM2LONG(val);
@@ -1451,6 +1715,29 @@ rb_num2ull(val)
 
 #endif  /* HAVE_LONG_LONG */
 
+
+/*
+ * Document-class: Integer
+ *
+ *  <code>Integer</code> is the basis for the two concrete classes that
+ *  hold whole numbers, <code>Bignum</code> and <code>Fixnum</code>.
+ *     
+ */
+
+
+/*
+ *  call-seq:
+ *     int.to_i      => int
+ *     int.to_int    => int
+ *     int.floor     => int
+ *     int.ceil      => int
+ *     int.round     => int
+ *     int.truncate  => int
+ *
+ *  As <i>int</i> is already an <code>Integer</code>, all these
+ *  methods simply return the receiver.
+ */
+
 static VALUE
 int_to_i(num)
     VALUE num;
@@ -1458,6 +1745,13 @@ int_to_i(num)
     return num;
 }
 
+/*
+ *  call-seq:
+ *     int.integer? -> true
+ *  
+ *  Always returns <code>true</code>.
+ */
+
 static VALUE
 int_int_p(num)
     VALUE num;
@@ -1465,6 +1759,17 @@ int_int_p(num)
     return Qtrue;
 }
 
+/*
+ *  call-seq:
+ *     int.next    => integer
+ *     int.succ    => integer
+ *  
+ *  Returns the <code>Integer</code> equal to <i>int</i> + 1.
+ *     
+ *     1.next      #=> 2
+ *     (-1).next   #=> 0
+ */
+
 static VALUE
 int_succ(num)
     VALUE num;
@@ -1476,6 +1781,18 @@ int_succ(num)
     return rb_funcall(num, '+', 1, INT2FIX(1));
 }
 
+/*
+ *  call-seq:
+ *     int.chr    => string
+ *  
+ *  Returns a string containing the ASCII character represented by the
+ *  receiver's value.
+ *     
+ *     65.chr    #=> "A"
+ *     ?a.chr    #=> "a"
+ *     230.chr   #=> "\346"
+ */
+
 static VALUE
 int_chr(num)
     VALUE num;
@@ -1642,10 +1959,6 @@ fix_to_s(argc, argv, x)
     if (argc == 0) base = 10;
     else base = NUM2INT(b);
 
-    if (base == 2) {
-	/* rb_fix2str() does not handle binary */
-	return rb_big2str(rb_int2big(FIX2INT(x)), 2);
-    }
     return rb_fix2str(x, base);
 }
 
@@ -1893,6 +2206,9 @@ fix_pow(x, y)
 	    return rb_big_pow(rb_int2big(a), y);
 	}
 	return rb_float_new(pow((double)a, (double)b));
+    } else if (TYPE(y) == T_FLOAT) {
+        long a = FIX2LONG(x);
+        return rb_float_new(pow((double)a, RFLOAT(y)->value));
     }
     return rb_num_coerce_bin(x, y);
 }
@@ -2145,7 +2461,7 @@ fix_lshift(x, y)
  * call-seq:
  *   fix >> count     => integer
  *
- * Shifts _fix_ left _count_ positions (right if _count_ is negative).
+ * Shifts _fix_ right _count_ positions (left if _count_ is negative).
  */
 
 static VALUE
@@ -2320,6 +2636,20 @@ fix_size(fix)
     return INT2FIX(sizeof(long));
 }
 
+/*
+ *  call-seq:
+ *     int.upto(limit) {|i| block }     => int
+ *  
+ *  Iterates <em>block</em>, passing in integer values from <i>int</i>
+ *  up to and including <i>limit</i>.
+ *     
+ *     5.upto(10) { |i| print i, " " }
+ *     
+ *  <em>produces:</em>
+ *     
+ *     5 6 7 8 9 10
+ */
+
 static VALUE
 int_upto(from, to)
     VALUE from, to;
@@ -2344,6 +2674,21 @@ int_upto(from, to)
     return from;
 }
 
+/*
+ *  call-seq:
+ *     int.downto(limit) {|i| block }     => int
+ *  
+ *  Iterates <em>block</em>, passing decreasing values from <i>int</i>
+ *  down to and including <i>limit</i>.
+ *     
+ *     5.downto(1) { |n| print n, ".. " }
+ *     print "  Liftoff!\n"
+ *     
+ *  <em>produces:</em>
+ *     
+ *     5.. 4.. 3.. 2.. 1..   Liftoff!
+ */
+
 static VALUE
 int_downto(from, to)
     VALUE from, to;
@@ -2368,6 +2713,22 @@ int_downto(from, to)
     return from;
 }
 
+/*
+ *  call-seq:
+ *     int.times {|i| block }     => int
+ *  
+ *  Iterates block <i>int</i> times, passing in values from zero to
+ *  <i>int</i> - 1.
+ *     
+ *     5.times do |i|
+ *       print i, " "
+ *     end
+ *     
+ *  <em>produces:</em>
+ *     
+ *     0 1 2 3 4
+ */
+
 static VALUE
 int_dotimes(num)
     VALUE num;
@@ -2419,6 +2780,9 @@ Init_Numeric()
 #elif defined(_UNICOSMP)
     /* Turn off floating point exceptions for divide by zero, etc. */
     _set_Creg(0, 0);
+#elif defined(__BORLANDC__)
+    /* Turn off floating point exceptions for overflow, etc. */
+    _control87(MCW_EM, MCW_EM);
 #endif
     id_coerce = rb_intern("coerce");
     id_to_i = rb_intern("to_i");
