@@ -2,8 +2,8 @@
 
   util.c -
 
-  $Author: matz $
-  $Date: 2004/09/21 09:35:28 $
+  $Author: nobu $
+  $Date: 2006/07/18 01:55:15 $
   created at: Fri Mar 10 17:22:34 JST 1995
 
   Copyright (C) 1993-2003 Yukihiro Matsumoto
@@ -15,6 +15,8 @@
 #include <ctype.h>
 #include <stdio.h>
 #include <errno.h>
+#include <math.h>
+#include <float.h>
 
 #ifdef _WIN32
 #include "missing/file.h"
@@ -679,12 +681,11 @@ ruby_getcwd()
  *
  */
 
-#define TRUE 1
-#define FALSE 0
+#define MDMINEXPT DBL_MIN_EXP
+#define MDMAXEXPT DBL_MAX_EXP
 
-static  int     MDMINEXPT       = -323;
-static  int     MDMAXEXPT       =  309;
-static double powersOf10[] = {	/* Table giving binary powers of 10.  Entry */
+static const
+double powersOf10[] = { 	/* Table giving binary powers of 10.  Entry */
     10.0,			/* is 10^2^i.  Used to convert decimal */
     100.0,			/* exponents into floating-point numbers. */
     1.0e4,
@@ -735,8 +736,9 @@ ruby_strtod(string, endPtr)
     char **endPtr;		/* If non-NULL, store terminating character's
 				 * address here. */
 {
-    int sign, expSign = FALSE;
-    double fraction, dblExp, *d;
+    int sign, expSign = Qfalse;
+    double fraction = 0.0, dblExp;
+    const double *d;
     register const char *p;
     register int c;
     int exp = 0;		/* Exponent read from "EX" field. */
@@ -750,8 +752,8 @@ ruby_strtod(string, endPtr)
 				 * case, fracExp is incremented one for each
 				 * dropped digit. */
     int mantSize = 0;		/* Number of digits in mantissa. */
-    int hasPoint = FALSE;	/* Decimal point exists. */
-    int hasDigit = FALSE;	/* I or F exists. */
+    int hasPoint = Qfalse;	/* Decimal point exists. */
+    int hasDigit = Qfalse;	/* I or F exists. */
     const char *pMant;		/* Temporarily holds location of mantissa
 				 * in string. */
     const char *pExp;		/* Temporarily holds location of exponent
@@ -763,44 +765,43 @@ ruby_strtod(string, endPtr)
 
     errno = 0;
     p = string;
-    while (ISSPACE(*p)) {
-	p += 1;
-    }
+    while (ISSPACE(*p)) p++;
     if (*p == '-') {
-	sign = TRUE;
-	p += 1;
+	sign = Qtrue;
+	p++;
     }
     else {
-	if (*p == '+') {
-	    p += 1;
-	}
-	sign = FALSE;
+	if (*p == '+') p++;
+	sign = Qfalse;
     }
 
+    fraction = 0.;
+    exp = 0;
+
     /*
      * Count the number of digits in the mantissa
      * and also locate the decimal point.
      */
 
-    for ( ; c = *p; p += 1) {
+    for ( ; (c = *p) != '\0'; p++) {
 	if (!ISDIGIT(c)) {
-	    if (c != '.' || hasPoint) {
+	    if (c != '.' || hasPoint || !ISDIGIT(p[1])) {
 		break;
 	    }
-	    hasPoint = TRUE;
+	    hasPoint = Qtrue;
 	}
 	else {
 	    if (hasPoint) { /* already in fractional part */
-		fracExp -= 1;
+		fracExp--;
 	    }
 	    if (mantSize) { /* already in mantissa */
-		mantSize += 1;
+		mantSize++;
 	    }
 	    else if (c != '0') { /* have entered mantissa */
-		mantSize += 1;
+		mantSize++;
 		pMant = p;
 	    }
-	    hasDigit = TRUE;
+	    hasDigit = Qtrue;
 	}
     }
 
@@ -811,7 +812,7 @@ ruby_strtod(string, endPtr)
      * they can't affect the value anyway.
      */
     
-    pExp  = p;
+    pExp = p;
     if (mantSize) {
 	p = pMant;
     }
@@ -824,7 +825,7 @@ ruby_strtod(string, endPtr)
 	p = string;
     }
     else {
-	int frac1, frac2;
+	double frac1, frac2;
 	frac1 = 0;
 	for ( ; mantSize > 9; mantSize -= 1) {
 	    c = *p;
@@ -852,20 +853,25 @@ ruby_strtod(string, endPtr)
 
 	p = pExp;
 	if ((*p == 'E') || (*p == 'e')) {
-	    p += 1;
+	    p++;
 	    if (*p == '-') {
-		expSign = TRUE;
-		p += 1;
+		expSign = Qtrue;
+		p++;
 	    }
 	    else {
 		if (*p == '+') {
-		    p += 1;
+		    p++;
 		}
-		expSign = FALSE;
+		expSign = Qfalse;
 	    }
-	    while (ISDIGIT(*p)) {
-		exp = exp * 10 + (*p - '0');
-		p += 1;
+	    if (ISDIGIT(*p)) {
+		do {
+		    exp = exp * 10 + (*p++ - '0');
+		}
+		while (ISDIGIT(*p));
+	    }
+	    else {
+		p = pExp;
 	    }
 	}
 	if (expSign) {
@@ -882,22 +888,24 @@ ruby_strtod(string, endPtr)
 	 * fraction.
 	 */
     
-	if (exp >= MDMAXEXPT - 18) {
-	    exp = MDMAXEXPT;
+	if (exp >= MDMAXEXPT) {
 	    errno = ERANGE;
+	    fraction = HUGE_VAL;
+	    goto ret;
 	}
-	else if (exp < MDMINEXPT + 18) {
-	    exp = MDMINEXPT;
+	else if (exp < MDMINEXPT) {
 	    errno = ERANGE;
+	    fraction = 0.0;
+	    goto ret;
 	}
 	fracExp = exp;
 	exp += 9;
 	if (exp < 0) {
-	    expSign = TRUE;
+	    expSign = Qtrue;
 	    exp = -exp;
 	}
 	else {
-	    expSign = FALSE;
+	    expSign = Qfalse;
 	}
 	dblExp = 1.0;
 	for (d = powersOf10; exp != 0; exp >>= 1, d += 1) {
@@ -906,18 +914,18 @@ ruby_strtod(string, endPtr)
 	    }
 	}
 	if (expSign) {
-	    fraction = frac1 / dblExp;
+	    frac1 /= dblExp;
 	}
 	else {
-	    fraction = frac1 * dblExp;
+	    frac1 *= dblExp;
 	}
 	exp = fracExp;
 	if (exp < 0) {
-	    expSign = TRUE;
+	    expSign = Qtrue;
 	    exp = -exp;
 	}
 	else {
-	    expSign = FALSE;
+	    expSign = Qfalse;
 	}
 	dblExp = 1.0;
 	for (d = powersOf10; exp != 0; exp >>= 1, d += 1) {
@@ -926,17 +934,18 @@ ruby_strtod(string, endPtr)
 	    }
 	}
 	if (expSign) {
-	    fraction += frac2 / dblExp;
+	    frac2 /= dblExp;
 	}
 	else {
-	    fraction += frac2 * dblExp;
+	    frac2 *= dblExp;
 	}
+	fraction = frac1 + frac2;
     }
 
+  ret:
     if (endPtr != NULL) {
-	*endPtr = (char *) p;
+	*endPtr = (char *)p;
     }
-
     if (sign) {
 	return -fraction;
     }
